#!/usr/bin/env python
"""
A wrapper around pytest that attempts to ensure correct arguments are passed to
it both when running locally and through cibuildhweel.
"""
from os.path import dirname, join, abspath
import sqlite3
import sys
import os
import re


def is_cibuildwheel():
    """Check if run with cibuildwheel."""
    return 'CIBUILDWHEEL' in os.environ


def temp_rename_kernprof(repo_dir):
    """
    Hacky workaround so kernprof.py doesn't get covered twice (installed and local).
    This needed to combine the .coverage files, since file paths need to be unique.

    """
    original_path = repo_dir + '/kernprof.py'
    tmp_path = original_path + '.tmp'
    if os.path.isfile(original_path):
        os.rename(original_path, tmp_path)
    elif os.path.isfile(tmp_path):
        os.rename(tmp_path, original_path)


def replace_docker_path(path, runner_project_dir):
    """Update path to a file installed in a temp venv to runner_project_dir."""
    pattern = re.compile(r"\/tmp\/.+?\/site-packages")
    return pattern.sub(runner_project_dir, path)


def update_coverage_file(coverage_path, runner_project_dir):
    """
    Since the paths inside of docker vary from the runner paths,
    the paths in the .coverage file need to be adjusted to combine them,
    since 'coverage combine <folder>' checks if the file paths exist.
    """
    try:
        sqliteConnection = sqlite3.connect(coverage_path)
        cursor = sqliteConnection.cursor()
        print('Connected to Coverage SQLite')

        read_file_query = 'SELECT id, path from file'
        cursor.execute(read_file_query)

        old_records = cursor.fetchall()
        new_records = [(replace_docker_path(path, runner_project_dir), _id) for _id, path in old_records]
        print('Updated coverage file paths:\n', new_records)

        sql_update_query = 'Update file set path = ? where id = ?'
        cursor.executemany(sql_update_query, new_records)
        sqliteConnection.commit()
        print('Coverage Updated successfully')
        cursor.close()

    except sqlite3.Error as error:
        print('Failed to coverage: ', error)
    finally:
        if sqliteConnection:
            sqliteConnection.close()
            print('The sqlite connection is closed')


def copy_coverage_cibuildwheel_docker(runner_project_dir):
    """
    When run with cibuildwheel under linux, the tests run in the folder /project
    inside docker and the coverage files need to be copied to the output folder.
    """
    coverage_path = '/project/tests/.coverage'
    if os.path.isfile(coverage_path):
        update_coverage_file(coverage_path, runner_project_dir)
        env_hash = hash((sys.version, os.environ.get('AUDITWHEEL_PLAT', '')))
        os.makedirs('/output', exist_ok=True)
        os.rename(coverage_path, '/output/.coverage.{}'.format(env_hash))


if __name__ == '__main__':
    cwd = os.getcwd()
    repo_dir = abspath(dirname(__file__))
    test_dir = join(repo_dir, 'tests')
    print('cwd = {!r}'.format(cwd))

    import pytest

    # Prefer testing the installed version, but fallback to testing the
    # development version.
    try:
        import ubelt as ub
    except ImportError:
        print('running this test script requires ubelt')
        raise
    package_name = 'pycold'
    # Statically check if ``package_name`` is installed outside of the repo.
    # To do this, we make a copy of PYTHONPATH, remove the repodir, and use
    # ubelt to check to see if ``package_name`` can be resolved to a path.
    temp_path = list(map(abspath, sys.path))
    if repo_dir in temp_path:
        temp_path.remove(repo_dir)
    modpath = ub.modname_to_modpath(package_name, sys_path=temp_path)
    if modpath is not None:
        # If it does, then import it. This should cause the installed version
        # to be used on further imports even if the repo_dir is in the path.
        print(f'Using installed version of {package_name}')
        module = ub.import_module_from_path(modpath, index=0)
        print('Installed module = {!r}'.format(module))
    else:
        print(f'No installed version of {package_name} found')

    try:
        print('Changing dirs to test_dir={!r}'.format(test_dir))
        os.chdir(test_dir)

        pytest_args = [
            '--cov-config', '../pyproject.toml',
            '--cov-report', 'html',
            '--cov-report', 'term',
            '--cov-report', 'xml',
            '--cov=' + package_name,
            modpath, '.'
        ]
        if is_cibuildwheel():
            pytest_args.append('--cov-append')

        pytest_args = pytest_args + sys.argv[1:]
        sys.exit(pytest.main(pytest_args))
    finally:
        os.chdir(cwd)
        if is_cibuildwheel():
            # for CIBW under linux
            copy_coverage_cibuildwheel_docker(f'/home/runner/work/{package_name}/{package_name}')
        print('Restoring cwd = {!r}'.format(cwd))



from libc.stdlib cimport malloc
from libc.string cimport strcpy, strlen
import numpy as np
# "cimport" is used to import special compile-time information
# about the numpy module (this is stored in a file numpy.pxd which is
# currently part of the Cython distribution).
cimport numpy as np
import math
from libcpp cimport bool
from cpython.mem cimport PyMem_Malloc, PyMem_Realloc, PyMem_Free
from collections import namedtuple
from copy import deepcopy


try:
    import typing
    import dataclasses
except ImportError:
    pass  # The modules don't actually have to exists for Cython to use them as annotations

cdef int NUM_FC = 40  # define the maximum number of outputted curves
cdef int NUM_FC_SCCD = 20
cdef int NUM_NRT_QUEUE = 240
DEF DEFAULT_CONSE = 6
DEF NRT_BAND = 6
DEF SCCD_NUM_C = 6

reccg_dt = np.dtype([('t_start', np.int32),  # time when series model gets started
                     ('t_end', np.int32),  # time when series model gets ended
                     ('t_break', np.int32),  # time when the first break (change) is observed
                     ('pos', np.int32),  # the location of each time series model
                     ('num_obs', np.int32),  # the number of "good" observations used for model estimation
                     ('category', np.short),  # the quality of the model estimation (what model is used, what process is used)
                     ('change_prob', np.short),  # the probability of a pixel that have undergone change (between 0 and 100)
                     ('coefs', np.float32, (7, 8)),  # coefficients for each time series model for each spectral band
                     ('rmse', np.float32, 7),  # RMSE for each time series model for each spectral band
                     ('magnitude', np.float32, 7)])  # the magnitude of change difference between model prediction
                                          # and observation for each spectral band)

sccd_dt = np.dtype([('t_start', np.int32),
                        ('t_break', np.int32),
                        ('num_obs', np.int32),
                        ('coefs', np.float32, (SCCD_NUM_C, SCCD_NUM_C)),
                        ('rmse', np.float32, NRT_BAND),
                        ('magnitude', np.float32, NRT_BAND)],
                        align=True)

nrtqueue_dt = np.dtype([('clry', np.short, NRT_BAND), ('clrx_since1982', np.short)], align=True)
nrtmodel_dt = np.dtype([('t_start_since1982', np.short), ('num_obs', np.short), ('obs', np.short, (NRT_BAND, DEFAULT_CONSE)),
                         ('obs_date_since1982', np.short, DEFAULT_CONSE), ('covariance', np.float32, (NRT_BAND, 36)),
                         ('nrt_coefs', np.float32, (6, 6)), ('H', np.float32, NRT_BAND), ('rmse_sum', np.uint32, 6),
                         ('norm_cm', np.short), ('cm_angle', np.short), ('conse_last', np.ubyte)], align=True)


cdef extern from "../../cxx/output.h":
    ctypedef struct Output_t:
        int t_start
        int t_end
        int t_break
        int pos
        int num_obs
        short int category
        short int change_prob
        float coefs[7][8]
        float rmse[7]
        float magnitude[7]

cdef extern from "../../cxx/output.h":
    ctypedef struct Output_sccd:
        int t_start
        int t_break
        int num_obs
        float coefs[NRT_BAND][SCCD_NUM_C]
        float rmse[NRT_BAND]
        float magnitude[NRT_BAND]

cdef extern from "../../cxx/output.h":
    ctypedef struct output_nrtqueue:
        short int clry[NRT_BAND]
        short int clrx_since1982
        
cdef extern from "../../cxx/output.h":
    ctypedef struct output_nrtmodel:
        short int t_start_since1982
        short int num_obs
        short int obs[NRT_BAND][DEFAULT_CONSE]
        short int obs_date_since1982[DEFAULT_CONSE]
        float covariance[NRT_BAND][36]
        float nrt_coefs[NRT_BAND][SCCD_NUM_C]
        float H[NRT_BAND]
        unsigned int rmse_sum[NRT_BAND]
        short int norm_cm;
        short int cm_angle;
        unsigned char conse_last;

cdef extern from "../../cxx/output.h":
    ctypedef struct Output_sccd_pinpoint:
        int t_break
        float coefs[NRT_BAND][SCCD_NUM_C]
        short int obs[NRT_BAND][DEFAULT_CONSE]
        short int obs_date_since1982[DEFAULT_CONSE]
        short int norm_cm[DEFAULT_CONSE]
        short int cm_angle[DEFAULT_CONSE]
 
cdef Output_sccd t
cdef output_nrtqueue t2
cdef output_nrtmodel t3
cdef Output_sccd_pinpoint t4


cdef extern from "../../cxx/cold.h":
    cdef int cold(long *buf_b, long *buf_g, long *buf_r, long *buf_n, long *buf_s1, long *buf_s2,
                  long *buf_t, long *fmask_buf, long *valid_date_array, int valid_num_scenes, int pos, 
                  double tcg, int conse, bool b_output_cm, int starting_date, bool b_c2, Output_t *rec_cg,
                  int *num_fc, int cm_output_interval, short int *cm_outputs,
                  short int *cm_outputs_date);


cdef extern from "../../cxx/cold.h":
    cdef int obcold_reconstruction_procedure(long *buf_b, long *buf_g, long *buf_r, long *buf_n, long *buf_s1,
    long *buf_s2, long *buf_t,  long *fmask_buf, long *valid_date_array, int valid_num_scenes, long *break_dates,
    int break_date_len, int pos, bool b_c2, int conse, Output_t *rec_cg, int *num_fc)



cdef extern from "../../cxx/s_ccd.h":
    cdef int sccd(long *buf_b, long *buf_g, long *buf_r, long *buf_n, long *buf_s1, long *buf_s2, long *buf_t,
                  long *fmask_buf, long *valid_date_array, int valid_num_scenes, double tcg, int *num_fc, int *nrt_mode,
                  Output_sccd *rec_cg, output_nrtmodel *nrt_model, int *num_nrt_queue, output_nrtqueue *nrt_queue,
                  short int *min_rmse, int conse, bool b_c2, bool b_pinpoint, Output_sccd_pinpoint *rec_cg_pinpoint, 
                  int *num_fc_pinpoint, double gate_tcg)



#cdef class SccdOutput:
#    cdef public int position
#    cdef public np.ndarray rec_cg
#    cdef public int nrt_mode
#    cdef public tuple nrt_model
#    cdef public np.ndarray nrt_queue
#    def __init__(self, position, rec_cg, nrt_mode, nrt_model, nrt_queue):
#        self.position = position
#        self.rec_cg = rec_cg
#        self.nrt_mode = nrt_mode
#        self.nrt_model = nrt_model
#        self.nrt_queue = nrt_queue
SccdOutput = namedtuple("SccdOutput", "position rec_cg min_rmse nrt_mode nrt_model nrt_queue")
SccdReccgPinpoint = namedtuple("SccdReccgPinpoint", "position rec_cg_pinpoint")

def test_func():
    """
    This is a docstring
    """
    return None

def cold_detect(np.ndarray[np.int64_t, ndim=1] dates, np.ndarray[np.int64_t, ndim=1] ts_b, np.ndarray[np.int64_t, ndim=1] ts_g,
                np.ndarray[np.int64_t, ndim=1] ts_r, np.ndarray[np.int64_t, ndim=1] ts_n, np.ndarray[np.int64_t, ndim=1] ts_s1,
                np.ndarray[np.int64_t, ndim=1] ts_s2, np.ndarray[np.int64_t, ndim=1] ts_t, np.ndarray[np.int64_t, ndim=1] qas,
                double t_cg = 15.0863, int pos=1, int conse=6, bint b_output_cm=False,
                int starting_date=0, int n_cm=0, int cm_output_interval=0, bint b_c2=False):
    """
    Helper function to do COLD algorithm.

    	Parameters
    	----------
    	dates: 1d array of shape(observation numbers), list of ordinal dates
    	ts_b: 1d array of shape(observation numbers), time series of blue band.
    	ts_g: 1d array of shape(observation numbers), time series of green band
    	ts_r: 1d array of shape(observation numbers), time series of red band
    	ts_n: 1d array of shape(observation numbers), time series of nir band
   	    ts_s1: 1d array of shape(observation numbers), time series of swir1 band
    	ts_s2: 1d array of shape(observation numbers), time series of swir2 band
    	ts_t: 1d array of shape(observation numbers), time series of thermal band
    	qas: 1d array, the QA cfmask bands. '0' - clear; '1' - water; '2' - shadow; '3' - snow; '4' - cloud
    	t_cg: threshold of change magnitude, default is chi2.ppf(0.99,5)
         pos: position id of the pixel
    	conse: consecutive observation number
    	b_output_cm: bool, 'True' means outputting change magnitude and change magnitude dates, only for object-based COLD
    	starting_date: the starting date of the whole dataset to enable reconstruct CM_date,
                   	all pixels for a tile should have the same date, only for b_output_cm is True
        b_c2: bool, a temporal parameter to indicate if collection 2. C2 needs ignoring thermal band for valid pixel test due to the current low quality
    	cm_output_interval: the temporal interval of outputting change magnitudes
    	Note that passing 2-d array to c as 2-d pointer does not work, so have to pass separate bands

    	Returns
    	----------
    	change records: the COLD outputs that characterizes each temporal segment
    """

    cdef int valid_num_scenes = qas.shape[0]
    # allocate memory for rec_cg
    cdef int num_fc = 0
    cdef Output_t t
    cdef Output_t* rec_cg = <Output_t*> PyMem_Malloc(NUM_FC * sizeof(t))

    # make sure it is c contiguous array
    if dates.flags['C_CONTIGUOUS'] == False:
        dates = np.ascontiguousarray(dates)
    if ts_b.flags['C_CONTIGUOUS'] == False:
        ts_b = np.ascontiguousarray(ts_b)
    if ts_g.flags['C_CONTIGUOUS'] == False:
        ts_g = np.ascontiguousarray(ts_g)
    if ts_r.flags['C_CONTIGUOUS'] == False:
        ts_r = np.ascontiguousarray(ts_r)
    if ts_n.flags['C_CONTIGUOUS'] == False:
        ts_n = np.ascontiguousarray(ts_n)
    if ts_s1.flags['C_CONTIGUOUS'] == False:
        ts_s1 = np.ascontiguousarray(ts_s1)
    if ts_s2.flags['C_CONTIGUOUS'] == False:
        ts_s2 = np.ascontiguousarray(ts_s2)
    if ts_t.flags['C_CONTIGUOUS'] == False:
        ts_t = np.ascontiguousarray(ts_t)
    if qas.flags['C_CONTIGUOUS'] == False:
        qas = np.ascontiguousarray(qas)
    
    cdef long [:] dates_view = dates
    cdef long [:] ts_b_view = ts_b
    cdef long [:] ts_g_view = ts_g
    cdef long [:] ts_r_view = ts_r
    cdef long [:] ts_n_view = ts_n
    cdef long [:] ts_s1_view = ts_s1
    cdef long [:] ts_s2_view = ts_s2
    cdef long [:] ts_t_view = ts_t
    cdef long [:] qas_view = qas


    assert ts_b_view.shape[0] == dates_view.shape[0]
    assert ts_g_view.shape[0] == dates_view.shape[0]
    assert ts_r_view.shape[0] == dates_view.shape[0]
    assert ts_n_view.shape[0] == dates_view.shape[0]
    assert ts_s1_view.shape[0] == dates_view.shape[0]
    assert ts_s2_view.shape[0] == dates_view.shape[0]
    assert ts_t_view.shape[0] == dates_view.shape[0]
    assert qas_view.shape[0] == dates_view.shape[0]

    # cm_outputs and cm_outputs_date are for object-based cold
    if b_output_cm == True:
        if cm_output_interval == 0:
           cm_output_interval = 60
        if starting_date == 0:
           starting_date = dates[0]
        if n_cm == 0:
           n_cm = math.ceil((dates[valid_num_scenes-1] - starting_date + 1) / cm_output_interval) + 1
        cm_outputs = np.full(n_cm, -9999, dtype=np.short)
        cm_outputs_date = np.full(n_cm, -9999, dtype=np.short)
    # set the length to 1 to save memory, as they won't be assigned values
    else:  
        cm_outputs = np.full(1, -9999, dtype=np.short)
        cm_outputs_date = np.full(1, -9999, dtype=np.short)
    cdef short [:] cm_outputs_view = cm_outputs  # memory view
    cdef short [:] cm_outputs_date_view = cm_outputs_date  # memory view

    result = cold(&ts_b_view[0], &ts_g_view[0], &ts_r_view[0], &ts_n_view[0], &ts_s1_view[0], &ts_s2_view[0], &ts_t_view[0],
                 &qas_view[0], &dates_view[0], valid_num_scenes, pos, t_cg, conse, b_output_cm,
                 starting_date, b_c2, rec_cg, &num_fc, cm_output_interval, &cm_outputs_view[0], &cm_outputs_date_view[0])
    if result != 0:
        raise RuntimeError("cold function fails for pos = {} ".format(pos))
    else:
        if num_fc <= 0:
            raise Exception("The COLD function has no change records outputted for pos = {} (possibly due to no enough clear observation)".format(pos))
        else:
            if b_output_cm == False:
                return np.asarray(<Output_t[:num_fc]>rec_cg) # np.asarray uses also the buffer-protocol and is able to construct
                                                             # a dtype-object from cython's array
            else:  # for object-based COLD
                return [np.asarray(<Output_t[:num_fc]>rec_cg), cm_outputs, cm_outputs_date]


def obcold_reconstruct(np.ndarray[np.int64_t, ndim=1] dates,
                       np.ndarray[np.int64_t, ndim=1] ts_b,
                       np.ndarray[np.int64_t, ndim=1] ts_g,
                       np.ndarray[np.int64_t, ndim=1] ts_r,
                       np.ndarray[np.int64_t, ndim=1] ts_n,
                       np.ndarray[np.int64_t, ndim=1] ts_s1,
                       np.ndarray[np.int64_t, ndim=1] ts_s2,
                       np.ndarray[np.int64_t, ndim=1] ts_t,
                       np.ndarray[np.int64_t, ndim=1] qas,
                       np.ndarray[np.int64_t, ndim=1] break_dates, int pos=1,
                       int conse=6, bint b_c2=False):
    """
    Helper function to do COLD algorithm.

    	Parameters
    	----------
    	dates: 1d array of shape(observation numbers), list of ordinal dates
    	ts_b: 1d array of shape(observation numbers), time series of blue band.
    	ts_g: 1d array of shape(observation numbers), time series of green band
    	ts_r: 1d array of shape(observation numbers), time series of red band
    	ts_n: 1d array of shape(observation numbers), time series of nir band
   	    ts_s1: 1d array of shape(observation numbers), time series of swir1 band
    	ts_s2: 1d array of shape(observation numbers), time series of swir2 band
    	ts_t: 1d array of shape(observation numbers), time series of thermal band
    	qas: 1d array, the QA cfmask bands. '0' - clear; '1' - water; '2' - shadow; '3' - snow; '4' - cloud
    	break_dates: 1d array, the break dates obtained from other procedures such as obia
    	conse: consecutive observation number (for calculating change magnitudes)
    	b_c2: bool, a temporal parameter to indicate if collection 2. C2 needs ignoring thermal band for valid pixel test due to its current low quality
    	Note that passing 2-d array to c as 2-d pointer does not work, so have to pass separate bands

    	Returns
    	----------
    	change records: the COLD outputs that characterizes each temporal segment
    """

    cdef int valid_num_scenes = qas.shape[0]
    cdef int break_date_len = break_dates.shape[0]
    # allocate memory for rec_cg
    cdef int num_fc = 0
    cdef Output_t t
    cdef Output_t* rec_cg = <Output_t*> PyMem_Malloc(NUM_FC * sizeof(t))

    cdef long [:] dates_view = dates
    cdef long [:] ts_b_view = ts_b
    cdef long [:] ts_g_view = ts_g
    cdef long [:] ts_r_view = ts_r
    cdef long [:] ts_n_view = ts_n
    cdef long [:] ts_s1_view = ts_s1
    cdef long [:] ts_s2_view = ts_s2
    cdef long [:] ts_t_view = ts_t
    cdef long [:] qas_view = qas
    cdef long [:] break_dates_view = break_dates

    assert ts_b_view.shape[0] == dates_view.shape[0]
    assert ts_g_view.shape[0] == dates_view.shape[0]
    assert ts_r_view.shape[0] == dates_view.shape[0]
    assert ts_n_view.shape[0] == dates_view.shape[0]
    assert ts_s1_view.shape[0] == dates_view.shape[0]
    assert ts_s2_view.shape[0] == dates_view.shape[0]
    assert ts_t_view.shape[0] == dates_view.shape[0]
    assert qas_view.shape[0] == dates_view.shape[0]

    result = obcold_reconstruction_procedure(&ts_b_view[0], &ts_g_view[0], &ts_r_view[0], &ts_n_view[0], &ts_s1_view[0],
     &ts_s2_view[0], &ts_t_view[0], &qas_view[0], &dates_view[0], valid_num_scenes, &break_dates_view[0], break_date_len,
      pos, b_c2, conse, rec_cg, &num_fc)
    if result != 0:
        raise RuntimeError("cold function fails for pos = {} ".format(pos))
    else:
        if num_fc <= 0:
            raise Exception("The reconstruct function has no change records outputted for pos = {} (possibly due to no enough clear observation)".format(pos))
        else:
            return np.asarray(<Output_t[:num_fc]>rec_cg) # np.asarray uses also the buffer-protocol and is able to construct a dtype-object from cython's array


def sccd_detect(np.ndarray[np.int64_t, ndim=1] dates, np.ndarray[np.int64_t, ndim=1] ts_b, np.ndarray[np.int64_t, ndim=1] ts_g,
                np.ndarray[np.int64_t, ndim=1] ts_r, np.ndarray[np.int64_t, ndim=1] ts_n, np.ndarray[np.int64_t, ndim=1] ts_s1,
                np.ndarray[np.int64_t, ndim=1] ts_s2, np.ndarray[np.int64_t, ndim=1] ts_t, np.ndarray[np.int64_t, ndim=1] qas,
                double t_cg = 15.0863, int pos=1, int conse=6, bint b_c2=False, b_pinpoint=False, double gate_tcg=9.236):
    """
    S-CCD processing. It is required to be done before near real time monitoring

    	Parameters
    	----------
    	dates: 1d array of shape(observation numbers), list of ordinal dates
    	ts_b: 1d array of shape(observation numbers), time series of blue band.
    	ts_g: 1d array of shape(observation numbers), time series of green band
    	ts_r: 1d array of shape(observation numbers), time series of red band
    	ts_n: 1d array of shape(observation numbers), time series of nir band
   	    ts_s1: 1d array of shape(observation numbers), time series of swir1 band
    	ts_s2: 1d array of shape(observation numbers), time series of swir2 band
    	ts_t: 1d array of shape(observation numbers), time series of thermal band
    	qas: 1d array, the QA cfmask bands. '0' - clear; '1' - water; '2' - shadow; '3' - snow; '4' - cloud
    	t_cg: threshold of change magnitude, default is chi2.ppf(0.99,5)
        pos: position id of the pixel
    	conse: consecutive observation number
    	b_c2: bool, a temporal parameter to indicate if collection 2. C2 needs ignoring thermal band for valid pixel test due to its current low quality
    	b_pinpoint: bool, output pinpoint break
    	Note that passing 2-d array to c as 2-d pointer does not work, so have to pass separate bands
        
    	Returns
    	----------
        namedtupe: SccdOutput
            change records: the S-CCD outputs that characterizes each temporal segment
            rec_cg:
            min_rmse
            int nrt_mode,             /* O: 0 - void; 1 - monitor mode for standard; 2 - queue mode for standard; 3 - new change; 4 - monitor mode for snow; 5 - queue mode for snow */
            output_nrtmodel: nrt model if monitor mode, empty if queue mode
            output_nrtqueue: obs queue if queue mode, empty if monitor mode
    """
    if dates.flags['C_CONTIGUOUS'] == False:
       dates = np.ascontiguousarray(dates)
    if ts_b.flags['C_CONTIGUOUS'] == False:
        ts_b = np.ascontiguousarray(ts_b)
    if ts_g.flags['C_CONTIGUOUS'] == False:
        ts_g = np.ascontiguousarray(ts_g)
    if ts_r.flags['C_CONTIGUOUS'] == False:
        ts_r = np.ascontiguousarray(ts_r)
    if ts_n.flags['C_CONTIGUOUS'] == False:
        ts_n = np.ascontiguousarray(ts_n)
    if ts_s1.flags['C_CONTIGUOUS'] == False:
        ts_s1 = np.ascontiguousarray(ts_s1)
    if ts_s2.flags['C_CONTIGUOUS'] == False:
        ts_s2 = np.ascontiguousarray(ts_s2)
    if ts_t.flags['C_CONTIGUOUS'] == False:
        ts_t = np.ascontiguousarray(ts_t)
    if qas.flags['C_CONTIGUOUS'] == False:
        qas = np.ascontiguousarray(qas)
        
    if conse > DEFAULT_CONSE:
        raise RuntimeError("The inputted conse is longer than the maximum conse for S-CCD: {}".format(DEFAULT_CONSE))

    cdef int valid_num_scenes = qas.shape[0]
    # allocate memory for rec_cg
    cdef int num_fc = 0
    cdef int num_nrt_queue = 0
    cdef int num_fc_pinpoint = 0
    cdef Output_sccd* rec_cg = <Output_sccd*> PyMem_Malloc(NUM_FC_SCCD * sizeof(t))
    cdef Output_sccd_pinpoint* rec_cg_pinpoint = <Output_sccd_pinpoint*> PyMem_Malloc(NUM_FC_SCCD * sizeof(t4))
    cdef output_nrtqueue* nrt_queue = <output_nrtqueue*> PyMem_Malloc(NUM_NRT_QUEUE * sizeof(t2))
    cdef output_nrtmodel* nrt_model = <output_nrtmodel*> PyMem_Malloc(sizeof(t3))
    cdef int nrt_mode = 0
    # initiate minimum rmse
    min_rmse = np.full(NRT_BAND, 0, dtype=np.short)

    # memory view
    cdef long [:] dates_view = dates
    cdef long [:] ts_b_view = ts_b
    cdef long [:] ts_g_view = ts_g
    cdef long [:] ts_r_view = ts_r
    cdef long [:] ts_n_view = ts_n
    cdef long [:] ts_s1_view = ts_s1
    cdef long [:] ts_s2_view = ts_s2
    cdef long [:] ts_t_view = ts_t
    cdef long [:] qas_view = qas
    cdef short [:] min_rmse_view = min_rmse


    assert ts_b_view.shape[0] == dates_view.shape[0]
    assert ts_g_view.shape[0] == dates_view.shape[0]
    assert ts_r_view.shape[0] == dates_view.shape[0]
    assert ts_n_view.shape[0] == dates_view.shape[0]
    assert ts_s1_view.shape[0] == dates_view.shape[0]
    assert ts_s2_view.shape[0] == dates_view.shape[0]
    assert ts_t_view.shape[0] == dates_view.shape[0]
    assert qas_view.shape[0] == dates_view.shape[0]


    result = sccd(&ts_b_view[0], &ts_g_view[0], &ts_r_view[0], &ts_n_view[0], &ts_s1_view[0], &ts_s2_view[0],
                  &ts_t_view[0], &qas_view[0], &dates_view[0], valid_num_scenes, t_cg, &num_fc, &nrt_mode, rec_cg,
                  nrt_model, &num_nrt_queue, nrt_queue, &min_rmse_view[0], conse, b_c2, b_pinpoint, rec_cg_pinpoint,
                  &num_fc_pinpoint, gate_tcg)
    if result != 0:
        raise RuntimeError("S-CCD function fails for pos = {} ".format(pos))
    else:
        if nrt_mode < 0:
            raise RuntimeError("No correct nrt_mode returned for pos = {} ".format(pos))
        else:
            if num_fc > 0:
                output_rec_cg = np.asarray(<Output_sccd[:num_fc]>rec_cg)
            else:
                output_rec_cg = np.array([])
                
            if b_pinpoint == False:
                if nrt_mode == 1 or nrt_mode == 3:  # monitor mode
                    return SccdOutput(pos, output_rec_cg, min_rmse, nrt_mode,
                                      np.asarray(<output_nrtmodel[:1]>nrt_model), np.array([]))
                elif nrt_mode == 2 or nrt_mode == 4:  # queue mode
                    return SccdOutput(pos, output_rec_cg, min_rmse, nrt_mode,
                                      np.array([]),np.asarray(<output_nrtqueue[:num_nrt_queue]>nrt_queue))
                elif nrt_mode == 5:  # queue recent
                    return SccdOutput(pos, output_rec_cg, min_rmse, nrt_mode,
                                      np.asarray(<output_nrtmodel[:1]>nrt_model),
                                      np.asarray(<output_nrtqueue[:num_nrt_queue]>nrt_queue))
                elif nrt_mode == 0:  # void mode
                    return SccdOutput(pos, np.array([]), min_rmse, nrt_mode, np.array([]),
                                      np.array([]))
            else:
                if num_fc_pinpoint > 0:
                    output_rec_cg_pinpoint = np.asarray(<Output_sccd_pinpoint[:num_fc_pinpoint]>rec_cg_pinpoint)
                else:
                    output_rec_cg_pinpoint = np.array([])
           
                if nrt_mode == 1 or nrt_mode == 3:  # monitor mode
                    return [SccdOutput(pos, output_rec_cg, min_rmse, nrt_mode,
                                      np.asarray(<output_nrtmodel[:1]>nrt_model), np.array([])), 
                                      SccdReccgPinpoint(pos, output_rec_cg_pinpoint)]
                elif nrt_mode == 2 or nrt_mode == 4:  # queue mode
                    return [SccdOutput(pos, output_rec_cg, min_rmse, nrt_mode,
                                      np.array([]),np.asarray(<output_nrtqueue[:num_nrt_queue]>nrt_queue)), 
                                      SccdReccgPinpoint(pos, output_rec_cg_pinpoint)]
                elif nrt_mode == 5:  # queue recent
                    return [SccdOutput(pos, output_rec_cg, min_rmse, nrt_mode,
                                      np.asarray(<output_nrtmodel[:1]>nrt_model),
                                      np.asarray(<output_nrtqueue[:num_nrt_queue]>nrt_queue)), 
                                      SccdReccgPinpoint(pos, output_rec_cg_pinpoint)]
                elif nrt_mode == 0:  # void mode
                    return [SccdOutput(pos, np.array([]), min_rmse, nrt_mode, np.array([]),
                                      np.array([])), SccdReccgPinpoint(pos, output_rec_cg_pinpoint)]


def sccd_update(sccd_pack, np.ndarray[np.int64_t, ndim=1] dates, np.ndarray[np.int64_t, ndim=1] ts_b,
                np.ndarray[np.int64_t, ndim=1] ts_g, np.ndarray[np.int64_t, ndim=1] ts_r,
                np.ndarray[np.int64_t, ndim=1] ts_n, np.ndarray[np.int64_t, ndim=1] ts_s1,
                np.ndarray[np.int64_t, ndim=1] ts_s2, np.ndarray[np.int64_t, ndim=1] ts_t,
                np.ndarray[np.int64_t, ndim=1] qas, double t_cg = 15.0863, int pos=1, int conse=6, bint b_c2=False,
                double gate_tcg=9.236):
    """
    SCCD online update for new observations

       Parameters
       ----------
       sccd_pack: a namedtuple of SccdOutput
       dates: 1d array of shape(observation numbers), list of ordinal dates
       ts_b: 1d array of shape(observation numbers), time series of blue band.
       ts_g: 1d array of shape(observation numbers), time series of green band
       ts_r: 1d array of shape(observation numbers), time series of red band
       ts_n: 1d array of shape(observation numbers), time series of nir band
       ts_s1: 1d array of shape(observation numbers), time series of swir1 band
       ts_s2: 1d array of shape(observation numbers), time series of swir2 band
       ts_t: 1d array of shape(observation numbers), time series of thermal band
       qas: 1d array, the QA cfmask bands. '0' - clear; '1' - water; '2' - shadow; '3' - snow; '4' - cloud
       t_cg: threshold of change magnitude, default is chi2.ppf(0.99,5)
       pos: position id of the pixel
       conse: consecutive observation number
       b_c2: bool, a temporal parameter to indicate if collection 2. C2 needs ignoring thermal band for valid pixel test due to its current low quality
       Note that passing 2-d array to c as 2-d pointer does not work, so have to pass separate bands

       Returns
       ----------
       namedtupe: SccdOutput
            rec_cg: the S-CCD outputs that characterizes each temporal segment
            min_rmse
            int nrt_mode,             /* O: 0 - void; 1 - monitor mode for standard; 2 - queue mode for standard;
                                            3 - monitor mode for snow; 4 - queue mode for snow; 5 - new queue */
            nrt_model: nrt model if monitor mode, empty if queue mode
            nrt_queue: obs queue if queue mode, empty if monitor mode
    """
    if dates.flags['C_CONTIGUOUS'] == False:
       dates = np.ascontiguousarray(dates)
    if ts_b.flags['C_CONTIGUOUS'] == False:
        ts_b = np.ascontiguousarray(ts_b)
    if ts_g.flags['C_CONTIGUOUS'] == False:
        ts_g = np.ascontiguousarray(ts_g)
    if ts_r.flags['C_CONTIGUOUS'] == False:
        ts_r = np.ascontiguousarray(ts_r)
    if ts_n.flags['C_CONTIGUOUS'] == False:
        ts_n = np.ascontiguousarray(ts_n)
    if ts_s1.flags['C_CONTIGUOUS'] == False:
        ts_s1 = np.ascontiguousarray(ts_s1)
    if ts_s2.flags['C_CONTIGUOUS'] == False:
        ts_s2 = np.ascontiguousarray(ts_s2)
    if ts_t.flags['C_CONTIGUOUS'] == False:
        ts_t = np.ascontiguousarray(ts_t)
    if qas.flags['C_CONTIGUOUS'] == False:
        qas = np.ascontiguousarray(qas)

    # sccd_pack_copy = deepcopy(sccd_pack)

    cdef int valid_num_scenes = qas.shape[0]
    # allocate memory for rec_cg
    # cdef int num_fc = 0
    # cdef int num_nrt_queue = 0
    cdef int nrt_mode = sccd_pack.nrt_mode
    cdef int num_fc = len(sccd_pack.rec_cg)
    cdef int num_nrt_queue = len(sccd_pack.nrt_queue)
    cdef int num_fc_pinpoint = 0
    cdef Output_sccd_pinpoint* rec_cg_pinpoint = <Output_sccd_pinpoint*> PyMem_Malloc(sizeof(t4))

    # grab inputs from the input
    rec_cg_new = np.empty(NUM_FC_SCCD, dtype=sccd_dt)
    if num_fc > 0:
        rec_cg_new[0:num_fc] = sccd_pack.rec_cg[0:num_fc]
    # rec_cg = <Output_sccd*> PyMem_Malloc(NUM_FC * sizeof(t))

    nrt_queue_new = np.empty(NUM_NRT_QUEUE, dtype=nrtqueue_dt)
    if num_nrt_queue > 0:
        nrt_queue_new[0:num_nrt_queue] = sccd_pack.nrt_queue[0:num_nrt_queue]

    if nrt_mode == 1 or nrt_mode == 3 or nrt_mode == 5:
        nrt_model_new = sccd_pack.nrt_model.copy()
    else:
        nrt_model_new = np.empty(1, dtype=nrtmodel_dt)

    min_rmse = sccd_pack.min_rmse

    # memory view
    cdef Output_sccd [:] rec_cg_view = rec_cg_new
    cdef output_nrtqueue [:] nrt_queue_view = nrt_queue_new
    cdef output_nrtmodel [:] nrt_model_view = nrt_model_new
    cdef short [:] min_rmse_view = min_rmse
    cdef long [:] dates_view = dates
    cdef long [:] ts_b_view = ts_b
    cdef long [:] ts_g_view = ts_g
    cdef long [:] ts_r_view = ts_r
    cdef long [:] ts_n_view = ts_n
    cdef long [:] ts_s1_view = ts_s1
    cdef long [:] ts_s2_view = ts_s2
    cdef long [:] ts_t_view = ts_t
    cdef long [:] qas_view = qas

    # assert the length
    assert ts_b_view.shape[0] == dates_view.shape[0]
    assert ts_g_view.shape[0] == dates_view.shape[0]
    assert ts_r_view.shape[0] == dates_view.shape[0]
    assert ts_n_view.shape[0] == dates_view.shape[0]
    assert ts_s1_view.shape[0] == dates_view.shape[0]
    assert ts_s2_view.shape[0] == dates_view.shape[0]
    assert ts_t_view.shape[0] == dates_view.shape[0]
    assert qas_view.shape[0] == dates_view.shape[0]

    result = sccd(&ts_b_view[0], &ts_g_view[0], &ts_r_view[0], &ts_n_view[0], &ts_s1_view[0], &ts_s2_view[0],
                  &ts_t_view[0], &qas_view[0], &dates_view[0], valid_num_scenes, t_cg, &num_fc, &nrt_mode, &rec_cg_view[0],
                  &nrt_model_view[0], &num_nrt_queue, &nrt_queue_view[0], &min_rmse_view[0], conse, b_c2, False,
                  rec_cg_pinpoint, &num_fc_pinpoint, gate_tcg)
    if result != 0:
        raise RuntimeError("sccd_update function fails for pos = {} ".format(pos))
    else:
        if nrt_mode < 0:
            raise RuntimeError("No correct nrt_mode returned for pos = {} ".format(pos))
        else:
            # sccd_pack_copy = None
            if num_fc > 0:
                output_rec_cg = rec_cg_new[0:num_fc]
            else:
                output_rec_cg = np.array([])

            if nrt_mode == 1 or nrt_mode == 3:  # monitor mode
                return SccdOutput(pos, output_rec_cg, min_rmse, nrt_mode,
                                  nrt_model_new, np.array([]))
            elif nrt_mode == 2 or nrt_mode == 4:  # queue mode
                return SccdOutput(pos, output_rec_cg, min_rmse, nrt_mode,
                                  np.array([]), nrt_queue_new[0:num_nrt_queue])
            elif nrt_mode == 5:  # queue recent
                return SccdOutput(pos, output_rec_cg, min_rmse, nrt_mode,
                                  nrt_model_new, nrt_queue_new[0:num_nrt_queue])
            elif nrt_mode == 0:  # void mode
                return SccdOutput(pos, np.array([]), min_rmse, nrt_mode, np.array([]), np.array([]))



"""
This is a script to build a MWE of a build issue I'm having getting
scikit-build to properly install PyCold in development mode.

Usage:
    python setup.py mwe generate
    pip install -e .
    python setup.py mwe analyize
    pip uninstall pkg_mwe

    rm -rf _skbuild

    python setup.py mwe generate with_cxx && pip install -e . --verbose && python -c "from pkg_mwe.myalgo_cython import call_myalgo; print(call_myalgo())"

    python setup.py mwe analyize

    python ~/code/pycold/dev/setup.py mwe analyize --mod_name=pycold --repo_dpath=$HOME/code/pycold
"""
import sys


class ProjectStructure():
    def __init__(self, repo_dpath='.', mod_name='pkg_mwe'):
        import ubelt as ub
        self.root = ub.Path(repo_dpath)
        self.mod_name = mod_name
        self.cxx_path    = (self.root / 'src' / 'cxx')
        self.python_path = (self.root / 'src' / 'python')
        self.mod_dpath = (self.python_path / self.mod_name)

    def generate(self, with_cxx=0):
        self.python_path.delete().ensuredir()
        self.cxx_path.delete()
        (self.root / 'CMakeLists.txt').delete()
        self.mod_dpath.ensuredir()
        (self.mod_dpath / '__init__.py').write_text('__version__ = "1.0.0"')

        # Give the MWE a CXX extension
        WITH_CXX = with_cxx
        if WITH_CXX:
            self.cxx_path.ensuredir()
            import ubelt as ub

            (self.root / 'CMakeLists.txt').write_text(ub.codeblock(
                r'''
                cmake_minimum_required(VERSION 3.13.0)
                project(pkg_mwe LANGUAGES C Fortran)


                find_package(PythonInterp REQUIRED)
                find_package(PythonLibs REQUIRED)

                ###
                # Private helper function to execute `python -c "<cmd>"`
                #
                # Runs a python command and populates an outvar with the result of stdout.
                # Be careful of indentation if `cmd` is multiline.
                #
                function(pycmd outvar cmd)
                  execute_process(
                    COMMAND "${PYTHON_EXECUTABLE}" -c "${cmd}"
                    RESULT_VARIABLE _exitcode
                    OUTPUT_VARIABLE _output)
                  if(NOT ${_exitcode} EQUAL 0)
                    message(ERROR "Failed when running python code: \"\"\"
                ${cmd}\"\"\"")
                    message(FATAL_ERROR "Python command failed with error code: ${_exitcode}")
                  endif()
                  # Remove supurflous newlines (artifacts of print)
                  string(STRIP "${_output}" _output)
                  set(${outvar} "${_output}" PARENT_SCOPE)
                endfunction()

                ###
                # Find scikit-build and include its cmake resource scripts
                #
                if (NOT SKBUILD)
                  pycmd(skbuild_location "import os, skbuild; print(os.path.dirname(skbuild.__file__))")
                  set(skbuild_cmake_dir "${skbuild_location}/resources/cmake")
                  # If skbuild is not the driver, then we need to include its utilities in our CMAKE_MODULE_PATH
                  list(APPEND CMAKE_MODULE_PATH ${skbuild_cmake_dir})
                endif()

                find_package(PythonExtensions REQUIRED)
                find_package(Cython REQUIRED)
                find_package(NumPy REQUIRED)

                # Backend C library
                add_subdirectory("src/cxx")

                # Cython library
                add_subdirectory("src/python/pkg_mwe")
                '''))

            (self.cxx_path / 'myalgo.h').write_text(ub.codeblock(
                '''
                #ifndef MYALGO_H
                #define MYALGO_H
                int myalgo(long *arr1, long *arr2, size_t num);
                #endif MYALGO_H
                '''))
            (self.cxx_path / 'myalgo.c').write_text(ub.codeblock(
                r'''
                #include <string.h>
                long myalgo(long *arr1, long *arr2, size_t num)
                {
                    for (int i = 0 ; i < num ; i++ )
                    {
                        arr2[i] = arr1[i] + arr2[i];
                    }
                    return 1;
                }
                '''))
            cmake_list_cxx = self.cxx_path / 'CMakeLists.txt'
            cmake_list_cxx.write_text(ub.codeblock(
                '''
                set(MYALGO_MODULE_NAME "myalgo")
                list(APPEND MYALGO_SOURCES "myalgo.h" "myalgo.c")
                add_library(${MYALGO_MODULE_NAME} STATIC ${MYALGO_SOURCES})
                '''))

            (self.mod_dpath / 'myalgo_cython.pyx').write_text(ub.codeblock(
                '''
                import numpy as np
                cimport numpy as np
                cdef extern from "../../cxx/myalgo.h":
                    cdef int myalgo(long *arr1, long *arr2, size_t num);

                def call_myalgo():
                    """
                    This is a docstring
                    """
                    cdef int result;
                    cdef np.ndarray[np.int64_t, ndim=1] arr1
                    cdef np.ndarray[np.int64_t, ndim=1] arr2
                    arr1 = np.array([1, 2, 3], dtype=np.int64)
                    arr2 = np.array([4, 6, 9], dtype=np.int64)
                    cdef long [:] arr1_view = arr1
                    cdef long [:] arr2_view = arr2
                    cdef size_t num = len(arr1)
                    print(f'arr1={arr1}')
                    print(f'arr2={arr2}')
                    print('calling my algo')
                    result = myalgo(&arr1_view[0], &arr2_view[0], num)
                    print(f'arr1={arr1}')
                    print(f'arr2={arr2}')
                    return result
                '''))

            (self.mod_dpath / 'CMakeLists.txt').write_text(ub.codeblock(
                '''
                set(cython_source "myalgo_cython.pyx")
                set(PYMYALGO_MODULE_NAME "myalgo_cython")

                # Translate Cython into C/C++
                add_cython_target(${PYMYALGO_MODULE_NAME} "${cython_source}" C OUTPUT_VAR sources)

                # Add other C sources
                list(APPEND sources )

                # Create C++ library. Specify include dirs and link libs as normal
                add_library(${PYMYALGO_MODULE_NAME} MODULE ${sources})
                target_include_directories(
                    ${PYMYALGO_MODULE_NAME}
                    PUBLIC
                    ${NumPy_INCLUDE_DIRS}
                    ${PYTHON_INCLUDE_DIR}
                    ${CMAKE_CURRENT_SOURCE_DIR}
                )

                # TODO: not sure why this isn't set in the global scope?
                # Hack around it: just hard code the module name
                set(MYALGO_MODULE_NAME "myalgo")

                # TODO: linking to the MYALGO shared object isn't working 100% yet.
                target_link_libraries(${PYMYALGO_MODULE_NAME} ${MYALGO_MODULE_NAME})

                target_compile_definitions(${PYMYALGO_MODULE_NAME} PUBLIC
                    "NPY_NO_DEPRECATED_API"
                    #"NPY_1_7_API_VERSION=0x00000007"
                )

                # Transform the C++ library into an importable python module
                python_extension_module(${PYMYALGO_MODULE_NAME})

                # Install the C++ module to the correct relative location
                # (this will be an inplace build if you use `pip install -e`)
                #file(RELATIVE_PATH pymyalgo_install_dest "${CMAKE_SOURCE_DIR}" "${CMAKE_CURRENT_SOURCE_DIR}")

                # My "normal" method of setting install targets does not seem to work here. Hacking it.
                # NOTE: skbuild *seems* to place libraries in a data dir *unless* the install destination
                # corresponds exactly to the <package_dir>/<package_name> specified implicitly in setup.py
                set(pymyalgo_install_dest "src/python/pkg_mwe")
                #install(TARGETS ${MYALGO_MODULE_NAME} LIBRARY DESTINATION "${pymyalgo_install_dest}")
                install(TARGETS ${PYMYALGO_MODULE_NAME} LIBRARY DESTINATION "${pymyalgo_install_dest}")
                '''
            ))

    def analyize(self):
        from rich.console import Console
        from rich.panel import Panel
        from rich.syntax import Syntax
        from rich.table import Table
        import distutils.sysconfig
        import ubelt as ub
        import xdev

        console = Console()

        def rich_file_content(fpath, lexer='bash'):
            import os
            text = fpath.read_text()
            return Panel(Syntax(text, lexer), title=os.fspath(fpath))

        def print_egg_path_content(egg_info_dpath, color='blue'):
            blocklist = {'requires.txt'}
            fpaths = egg_info_dpath.ls()
            table = Table(f'[{color}]' + str(egg_info_dpath))
            for fpath in fpaths:
                if fpath.name not in blocklist:
                    panel = rich_file_content(fpath)
                    table.add_row(panel)
            console.print(table)

        print('\n')
        print('Repo Structure:')
        directory_blocklist = ['.*', '.git', 'dist', '_skbuild', 'dev']
        xdev.tree_repr(self.root, max_files=None, dirblocklist=directory_blocklist)

        print('\n')
        print('Content of the EGG Link:')
        site_dpath = ub.Path(distutils.sysconfig.get_python_lib())
        egg_link_fpaths = list(site_dpath.glob(self.mod_name.replace('_', '*') + '*.egg-link'))
        assert len(egg_link_fpaths) == 1
        egg_link_fpath = egg_link_fpaths[0]
        console.print(rich_file_content(egg_link_fpath))

        # Note: (recently 2022-08-ish) python switched to a new type of
        # This is not present in setuptools==63.2.0 but is in 65.3.0
        # editable install. TODO: incomporate this.
        # editable_fpaths = list(site_dpath.glob('__editable__*' + self.mod_name.replace('_', '*') + '*'))

        print('\n')
        print('Check easy-install.pth')
        easy_install_fpath = site_dpath / 'easy-install.pth'
        assert easy_install_fpath.exists()
        easy_install_text = easy_install_fpath.read_text()
        abs_path = self.mod_dpath.absolute().parent
        print(f'abs_path={abs_path}')
        if str(abs_path)  in easy_install_text:
            console.print('[green] Easy install dpath is good')
        else:
            console.print('[red] Easy install dpath is bad')
            console.print(rich_file_content(easy_install_fpath))

        expected_egg_info_dpath = self.root / f'src/python/{self.mod_name}.egg-info'
        all_egg_infos = [ub.Path(e).resolve() for e in xdev.find('*.egg-info', dirblocklist=directory_blocklist)]
        other_egg_infos = set(all_egg_infos) - {expected_egg_info_dpath.resolve()}
        print('expected_egg_info_dpath = {}'.format(ub.repr2(expected_egg_info_dpath, nl=1)))
        if expected_egg_info_dpath.exists():
            console.print('[green] Egg info exists in expected location')
            egg_info_dpath = expected_egg_info_dpath
            print_egg_path_content(egg_info_dpath, color='green')
        else:
            console.print('[red] Egg info exists in expected location')
            print(f'other_egg_infos={other_egg_infos}')

        if other_egg_infos:
            console.print('[red] THERE ARE UNEXEPCTED EGG INFOS')
            for egg_info_dpath in other_egg_infos:
                print_egg_path_content(egg_info_dpath, color='red')

        print('\n')
        print('Test to ensure we can import the module')
        command = f'python -c "import {self.mod_name}; print({self.mod_name})"'
        info = ub.cmd(command, verbose=3)
        if info['ret'] != 0:
            raise Exception('failed to import')
        assert str(self.mod_dpath) in info['out']


def mwe_cli():
    import scriptconfig as scfg
    class AnalyizeConfig(scfg.DataConfig):
        mod_name = 'pkg_mwe'
        repo_dpath = '.'
    config = AnalyizeConfig.cli()
    self = ProjectStructure(**config)
    if 'generate' in sys.argv:
        self.generate(with_cxx='with_cxx' in sys.argv)
    if 'analyize' in sys.argv:
        self.analyize()


if __name__ == '__main__':
    if 'mwe' in sys.argv:
        mwe_cli()
        sys.exit(0)

    import os
    DISABLE_C_EXTENSIONS = os.environ.get('DISABLE_C_EXTENSIONS', '')
    if DISABLE_C_EXTENSIONS == '1':
        from setuptools import setup
    else:
        from skbuild import setup
    from setuptools import find_packages

    packages = find_packages('./src/python')
    setup(
        package_dir={
            '': 'src/python',
        },
        install_requires=['packaging', 'ubelt'],
        name='pkg_mwe',
        version="1.0.0",
        description='MWE of build issue',
        packages=packages,
        include_package_data=True,
    )



# this script show how to run COLD from a csv file that store time series information
import numpy as np
import pathlib
from pycold import cold_detect, obcold_reconstruct
from pycold.utils import read_data

Landsat_bandname = ['Blue', 'Green', 'Red', 'NIR', 'SWIR1', 'SWIR2', 'Thermal']
t_c = -200  # the threshold used for get_breakcategory
w = np.pi * 2 / 365.25

# slope_in_rec_cg = original_slope * slope_scale, we used this way to keep precision for model coefficients in
# float datatype
slope_scale = 10000


# Use this file to determine where the resources are.  If for some reason we
# are do not have __file__ available (e.g.  copy/pasting in IPython) then
# assume we are in the repo reoot.
# TODO: should likely use pkg_resources instead
try:
    TEST_RESOURCE_DPATH = (pathlib.Path(__file__).parent / 'resources').resolve()
except NameError:
    TEST_RESOURCE_DPATH = pathlib.Path('tests/resources').resolve()


def get_breakcategory(ccd_plot, i_curve):
    """
    get break category:
    :param ccd_plot: rec_cg
    :param i_curve: the number of the curve to be analysised
    :return: 1 - disturbance break; 2 - natural recovery; 3 - aforestation
    see section 3.3.7 in Zhu, Z., Zhang, J., Yang, Z., Aljaddani, A. H., Cohen, W. B., Qiu, S., & Zhou, C. (2020).
    Continuous monitoring of land disturbance based on Landsat time series. Remote Sensing of Environment, 238, 111116.
    """
    if ccd_plot[i_curve]['magnitude'][3] > t_c and ccd_plot[i_curve]['magnitude'][2] < -t_c and \
            ccd_plot[i_curve]['magnitude'][4] < -t_c:
        if ccd_plot[i_curve + 1]['coefs'][3, 1] > np.abs(ccd_plot[i_curve]['coefs'][3, 1]) and \
                ccd_plot[i_curve + 1]['coefs'][2, 1] < -np.abs(ccd_plot[i_curve]['coefs'][2, 1]) and \
                ccd_plot[i_curve + 1]['coefs'][4, 1] < -np.abs(ccd_plot[i_curve]['coefs'][4, 1]):
            return 3
        else:
            return 2
    else:
        return 1


def test_cold_detect():
    # running COLD for a Landsat time series provided by a csv
    in_path = TEST_RESOURCE_DPATH / 'spectral_336_3980_obs.csv'
    data = read_data(in_path)
    dates, blues, greens, reds, nirs, swir1s, swir2s, thermals, qas, sensor = data.copy()  # exclude header
    cold_result = cold_detect(dates, blues, greens, reds, nirs, swir1s, swir2s, thermals, qas)
    assert cold_result[0]['t_break'] == 735040
    assert cold_result[0]['num_obs'] == 142
    assert cold_result[1]['t_end'] == 737368


def test_obcold_reconstruct():
    # running reconstructing for a Landsat time series provided by a csv
    in_path = TEST_RESOURCE_DPATH / 'spectral_336_3980_obs.csv'
    break_path = TEST_RESOURCE_DPATH / 'spectral_336_3980_breaks.csv'
    data = read_data(in_path)
    breaks = read_data(break_path)
    dates, blues, greens, reds, nirs, swir1s, swir2s, thermals, qas, sensor = data.copy()  # exclude header
    cold_result = obcold_reconstruct(dates, blues, greens, reds, nirs, swir1s, swir2s, thermals, qas, breaks)
    assert cold_result[0]['t_break'] == 735040
    # assert cold_result[0]['num_obs'] == 139
    # assert cold_result[1]['t_end'] == 737352



import yaml
import os
import shutil
import numpy as np
import joblib
import pathlib
from osgeo import gdal_array
from pycold.pyclassifier import PyClassifierHPC
from pycold.utils import assemble_array


# Use this file to determine where the resources are.  If for some reason we
# are do not have __file__ available (e.g.  copy/pasting in IPython) then
# assume we are in the repo reoot.
# TODO: should likely use pkg_resources instead
try:
    TEST_RESOURCE_DPATH = (pathlib.Path(__file__).parent / 'resources').resolve()
except NameError:
    TEST_RESOURCE_DPATH = pathlib.Path('tests/resources').resolve()


with open(TEST_RESOURCE_DPATH / 'test_config_pyclassifier.yaml', 'r') as yaml_obj:
    test_config = yaml.safe_load(yaml_obj)

ref_path = TEST_RESOURCE_DPATH / 'LCMAP_CU_2001_V01_LCPRI_027009_100_100.tif'
year_lowbound = 2015
year_uppbound = 2020
testing_folder = TEST_RESOURCE_DPATH / 'tmp'


def test_step1_4():
    pyclassifier = PyClassifierHPC(test_config,
                                   record_path=TEST_RESOURCE_DPATH,
                                   tmp_path=testing_folder,
                                   output_path=testing_folder,
                                   year_list_to_predict=list(range(year_lowbound, year_uppbound + 1)),
                                   seedmap_path=ref_path)

    # delete testing folder if it has
    if os.path.exists(pyclassifier.tmp_path):
        shutil.rmtree(pyclassifier.tmp_path)

    pyclassifier.hpc_preparation()

    for iblock in range(pyclassifier.config['n_blocks']):
        pyclassifier.step1_feature_generation(iblock + 1)
        for year in range(year_lowbound, year_uppbound + 1):
            assert (pathlib.Path(pyclassifier.tmp_path) /
                    'tmp_feature_year{}_block{}.npy'.format(year, iblock + 1)).exists()

    pyclassifier.step2_train_rf(2017)
    assert (pathlib.Path(pyclassifier.output_path) / 'rf.model').exists()

    for iblock in range(pyclassifier.config['n_blocks']):
        pyclassifier.step3_classification(iblock + 1)
        for year in range(year_lowbound, year_uppbound + 1):
            fname = 'tmp_yearlyclassification{}_block{}.npy'.format(year, iblock + 1)
            assert pathlib.Path(pyclassifier.tmp_path, fname).exists()
    pyclassifier.step4_assemble()
    for year in range(year_lowbound, year_uppbound + 1):
        fname = 'yearlyclassification_{}.npy'.format(year)
        assert pathlib.Path(pyclassifier.output_path, fname).exists()

    # import matplotlib.pyplot as plt
    # plt.imshow(np.load(TEST_RESOURCE_DPATH / 'tmp/yearlyclassification_2017.npy'))
    shutil.rmtree(pyclassifier.tmp_path)


def test_predict_features():
    pyclassifier = PyClassifierHPC(test_config, record_path=TEST_RESOURCE_DPATH,
                                   year_list_to_predict=list(range(year_lowbound, year_uppbound + 1)),
                                   seedmap_path=ref_path)
    cold_block = np.load(TEST_RESOURCE_DPATH / 'record_change_x1_y1_cold.npy')
    block_features = pyclassifier.predict_features(1, cold_block, list(range(year_lowbound, year_uppbound + 1)))
    assert np.shape(block_features) == (6, pyclassifier.config['block_width'] * pyclassifier.config['block_height'],
                                        21)   # 6 years, 2500 pixels (50, 50), 21 features


def test_train_rf_model():
    pyclassifier = PyClassifierHPC(test_config, record_path=TEST_RESOURCE_DPATH, tmp_path=TEST_RESOURCE_DPATH / 'feature_maps',
                                   year_list_to_predict=list(range(year_lowbound, year_lowbound + 1)),
                                   seedmap_path=ref_path)

    full_feature_2017 = assemble_array(pyclassifier.get_fullfeature_forcertainyear(2017),
                                       pyclassifier.config['n_block_x'])
    rf_model = pyclassifier.train_rfmodel(full_feature_2017, gdal_array.LoadFile(os.fspath(ref_path)))
    # pyclassifier.save_rf_model(rf_model)
    assert rf_model is not None


def test_classification_block():
    pyclassifier = PyClassifierHPC(test_config, record_path=TEST_RESOURCE_DPATH, tmp_path=TEST_RESOURCE_DPATH / 'feature_maps',
                                   rf_path=TEST_RESOURCE_DPATH / 'feature_maps/rf.model',
                                   year_list_to_predict=list(range(year_lowbound, year_lowbound + 1)))
    rf_model = joblib.load(TEST_RESOURCE_DPATH / 'feature_maps/rf.model')
    tmp_feature_block = np.load(TEST_RESOURCE_DPATH / 'feature_maps/tmp_feature_year2017_block1.npy')
    cmap = pyclassifier.classification_block(rf_model, tmp_feature_block)
    print(cmap.shape)
    assert cmap.shape == (50, 50)



import numpy as np
from datetime import datetime
import pandas as pd
from os.path import join
import os
import datetime as dt
from collections import namedtuple
from pycold.app import defaults
from osgeo import gdal

SccdOutput = namedtuple("SccdOutput", "position rec_cg min_rmse nrt_mode nrt_model nrt_queue")

sccd_dt = np.dtype([('t_start', np.int32), ('t_break', np.int32), ('num_obs', np.int32),
                    ('coefs', np.float32, (6, 6)), ('rmse', np.float32, 6), ('magnitude', np.float32, 6)], align=True)

nrtqueue_dt = np.dtype([('clry', np.short, 6), ('clrx_since1982', np.short)], align=True)
nrtmodel_dt = np.dtype([('t_start_since1982', np.short), ('num_obs', np.short), ('obs', np.short, (6, 6)),
                        ('obs_date_since1982', np.short, 6), ('covariance', np.float32, (6, 36)),
                        ('nrt_coefs', np.float32, (6, 6)), ('H', np.float32, 6), ('rmse_sum', np.uint32, 6),
                        ('norm_cm', np.short), ('cm_angle', np.short), ('conse_last', np.ubyte)], align=True)


def get_block_y(block_id, n_block_x):
    """
    Parameters
    ----------
    block_id: integer
    n_block_x: integer, number of blocks at x xis

    Returns
    -------
    current block id at y axis
    """
    return int((block_id - 1) / n_block_x) + 1


def get_block_x(block_id, n_block_x):
    """
    Parameters
    ----------
    block_id: integer
    n_block_x: integer, number of blocks at x xis

    Returns
    -------
    current block id at x axis
    """
    return (block_id - 1) % n_block_x + 1


def get_col_index(pos, n_cols, current_block_x, block_width):
    """
    Parameters
    ----------
    pos
    n_cols
    current_block_x
    block_width

    Returns
    -------

    """
    return int((pos - 1) % n_cols) - (current_block_x - 1) * block_width


def get_row_index(pos, n_cols, current_block_y, block_height):
    """
    Parameters
    ----------
    pos: start from 1
    n_cols
    current_block_y
    block_height

    Returns
    -------

    """
    return int((pos - 1) / n_cols) - (current_block_y - 1) * block_height


def assemble_cmmaps(config, result_path, cmmap_path, starting_date, n_cm_maps, prefix, clean=True):
    """
    this function reorganized block-based fix-interval CM intermediate files into map-based output (one map per interval)
    Parameters
    ----------
    config: dictionary
        pycold config dictionary
    result_path: string
        the path where block-based CM intermediate files are
    cmmap_path: string
        the path to save the new map-based output
    starting_date: integer
        the starting date of the dataset
    n_cm_maps: integer
        the number of change magnitude outputted per pixel
    prefix: {'CM', 'CM_date', 'CM_direction'}
    clean: True -> clean tmp files
    Returns
    -------

    """
    # anchor_dates_list = None
    if prefix == 'CM':
        output_type = np.int16
    elif prefix == 'CM_date':
        output_type = np.int32
        # cuz the date is produced as byte to squeeze the storage size, need to expand
        # anchor_dates_list_single = np.arange(start=starting_date,
        #                                      stop=starting_date + config['CM_OUTPUT_INTERVAL'] * n_cm_maps,
        #                                      step=config['CM_OUTPUT_INTERVAL'])
        # anchor_dates_list = np.tile(anchor_dates_list_single, config['block_width'] * config['block_height'])

    elif prefix == 'CM_direction':
        output_type = np.uint8

    cm_map_list = [np.full((config['n_rows'], config['n_cols']), defaults['COMMON']['NAN_VAL'], dtype=output_type) for x in range(n_cm_maps)]
    for iblock in range(config['n_blocks']):
        current_block_y = int(np.floor(iblock / config['n_block_x'])) + 1
        current_block_x = iblock % config['n_block_y'] + 1
        try:
            cm_block = np.load(join(result_path, '{}_x{}_y{}.npy'.format(prefix, current_block_x,
                                                                         current_block_y))).astype(output_type)
        except OSError as e:
            print('Reading CM files fails: {}'.format(e))
        #    continue

        if prefix == 'CM_date':
            cm_block_copy = cm_block.copy()
            cm_block = cm_block + defaults['COMMON']['JULIAN_LANDSAT4_LAUNCH']
            # we assign an extremely large value to original NA value (255)
            cm_block[cm_block_copy == -9999] = -9999

        cm_block_reshape = np.reshape(cm_block, (config['block_width'] * config['block_height'],
                                                 n_cm_maps))
        hori_profile = np.hsplit(cm_block_reshape, n_cm_maps)
        for count, maps in enumerate(cm_map_list):
            maps[(current_block_y - 1) * config['block_height']:current_block_y * config['block_height'],
                 (current_block_x - 1) * config['block_width']:current_block_x * config['block_width']] = (
                     hori_profile[count].reshape(config['block_height'], config['block_width'])
                 )

    # output cm images
    for count, cm_map in enumerate(cm_map_list):
        ordinal_date = starting_date + count * config['CM_OUTPUT_INTERVAL']
        outfile = join(cmmap_path, '{}_maps_{}_{}{}.npy'.format(prefix, str(ordinal_date),
                                                                pd.Timestamp.fromordinal(ordinal_date).year,
                                                                get_doy(ordinal_date)))
        np.save(outfile, cm_map)

    if clean is True:
        tmp_filenames = [file for file in os.listdir(result_path)
                         if file.startswith(prefix + '_x')]
        for file in tmp_filenames:
            os.remove(join(result_path, file))


def get_rowcol_intile(id, block_width, block_height, block_x, block_y):
    """
    calculate row and col in original images based on pos index and block location
    Parameters
    ----------
    id: integer
        position id of the pixel (i.e., i_row * n_cols + i_col)
    block_width: integer
        the width of each block
    block_height: integer
        the height of each block
    block_x:integer
        block location at x direction
    block_y:integer
        block location at y direction
    Returns
    -------
    (original_row, original_col)
    row and col number (starting from 1) in original image (e.g., Landsat ARD 5000*5000)
    """
    original_row = int(id / block_width + (block_y - 1) * block_height + 1)
    original_col = int(id % block_width + (block_x - 1) * block_width + 1)
    return original_row, original_col


def get_id_inblock(pos, block_width, block_height, n_cols):
    """
    :param pos:
    :param block_width:
    :param block_height:
    :param block_x:
    :param block_y:
    :return: pixel id in the bloack, starting from 0
    """
    row_inblock = int(int((pos - 1) / n_cols) % block_height)
    col_inblock = (pos - 1) % n_cols % block_width
    return row_inblock * block_width + col_inblock


# def gdal_save_file_1band(out_path, array, gdal_type, trans, proj, cols, rows, image_format='GTiff'):
#     """
#     save array as tiff format
#     Parameters
#     ----------
#     out_path : full outputted path
#     array : numpy array to be saved
#     gdal_type: gdal type
#     trans: transform coefficients
#     proj: projection
#     rows: the row number
#     cols: the col number
#     image_format: default is GTiff
#     Returns
#     -------
#     TRUE OR FALSE
#     """
#     outdriver = gdal.GetDriverByName(image_format)
#     outdata = outdriver.Create(out_path, cols, rows, 1, gdal_type)
#     if outdata == None:
#         return False
#     outdata.GetRasterBand(1).WriteArray(array)
#     outdata.FlushCache()
#     outdata.SetGeoTransform(trans)
#     outdata.FlushCache()
#     outdata.SetProjection(proj)
#     outdata.FlushCache()
#     return True


def get_time_now(tz):
    """
    Parameters
    ----------
    tz: string

    Returns
    -------
    datatime format of current time
    """
    return datetime.now(tz).strftime('%Y-%m-%d %H:%M:%S')


def get_ymd_now(tz):
    """
    Parameters
    ----------
    tz: string

    Returns
    -------
    datatime format of current time
    """
    return datetime.now(tz).strftime('%Y-%m-%d')


def get_doy(ordinal_date):
    """
    Parameters
    ----------
    ordinal_date: int
    a ordinal date (MATLAB-format ordinal date)

    Returns: string
    -------
    doy
    """
    return str(pd.Timestamp.fromordinal(ordinal_date).timetuple().tm_yday).zfill(3)


def get_anchor_days(starting_day, n_cm_maps, interval):
    """
    get a list of starting days for each change magnitude time slices
    Parameters
    ----------
    starting_days
    n_cm_maps
    interval

    Returns
    -------

    """
    return np.arange(start=starting_day, stop=starting_day + n_cm_maps * interval,
                     step=interval)


def assemble_array(array_list, n_block_x):
    """
    assemble a list of block-based array to a bigger array that aligns with the dimension of an ARD tile
    Parameters
    ----------
    array_list
    n_block_x: block number at x axis

    Returns
    -------
    an array [nrows, ncols]
    """
    full_feature_array = np.hstack(array_list)
    full_feature_array = np.vstack(np.hsplit(full_feature_array, n_block_x))  # (nrows, ncols, nfeatures)
    return full_feature_array


def read_blockdata(block_folder, total_pixels, total_bands):
    img_files = [f for f in os.listdir(block_folder) if f.startswith('L')]

    # sort image files by dates
    img_dates = [pd.Timestamp.toordinal(dt.datetime(int(folder_name[9:13]), 1, 1) +
                                        dt.timedelta(int(folder_name[13:16]) - 1))
                 for folder_name in img_files]
    files_date_zip = sorted(zip(img_dates, img_files))
    img_files_sorted = [x[1] for x in files_date_zip]
    img_dates_sorted = np.asarray([x[0] for x in files_date_zip])
    img_stack = [np.load(join(block_folder, f)).reshape(total_pixels, total_bands)
                 for f in img_files_sorted]
    img_stack = np.dstack(img_stack)
    return img_stack, img_dates_sorted


def read_data(path):
    """Load a sample file containing acquisition days and spectral values.
    The first column is assumed to be the day number, subsequent columns
    correspond to the day number. This improves readability of large datasets.
    Args:
        path: location of CSV containing test data
    Returns:
        A 2D numpy array.
    """
    return np.genfromtxt(path, delimiter=',', dtype=np.int64).T


def date2matordinal(year, month, day):
    return pd.Timestamp.toordinal(dt.date(year, month, day))


def matordinal2date(ordinal):
    return pd.Timestamp.fromordinal(ordinal)


def save_nrtfiles(out_folder, outfile_prefix, sccd_pack, data_ext):
    """
    save all files for C debug
    :param out_folder: the outputted folder
    :param outfile_prefix: the prefix of outputted files
    :param sccd_pack:
    :param data_ext:
    :return:
    """
    data_ext.to_csv(join(out_folder, 'spectral_{}_extension.csv').format(outfile_prefix), index=False, header=False)
    # data_ini_current.to_csv(join(out_path, 'spectral_{}_ini.csv').format(pid), index=False, header=False)
    np.asarray(sccd_pack.nrt_mode).tofile(join(out_folder, 'sccd_pack{}_nrt_mode').format(outfile_prefix))
    sccd_pack.rec_cg.tofile(join(out_folder, 'sccd_pack{}_rec_cg').format(outfile_prefix))
    sccd_pack.nrt_model.tofile(join(out_folder, 'sccd_pack{}_nrt_model').format(outfile_prefix))
    sccd_pack.nrt_queue.tofile(join(out_folder, 'sccd_pack{}_nrt_queue').format(outfile_prefix))
    sccd_pack.min_rmse.tofile(join(out_folder, 'sccd_pack{}_min_rmse').format(outfile_prefix))


def save_obs2csv(out_path, data):
    data.to_csv(out_path, index=False, header=False)


def unindex_sccdpack(sccd_pack_single):
    """
    remove index of sccdpack to save memory
    sccd_pack_single: a namedtuple SccdOutput
    :return: a nested list
    """
    sccd_pack_single = sccd_pack_single._replace(rec_cg=sccd_pack_single.rec_cg.tolist())

    if len(sccd_pack_single.nrt_model) > 0:
        sccd_pack_single = sccd_pack_single._replace(nrt_model=sccd_pack_single.nrt_model.tolist())

    if len(sccd_pack_single.nrt_queue) > 0:
        sccd_pack_single = sccd_pack_single._replace(nrt_queue=sccd_pack_single.nrt_queue.tolist())

    return list(sccd_pack_single)


def index_sccdpack(sccd_pack_single):
    """
    convert list of sccdpack to namedtuple to facilitate parse,
    :param sccd_pack_single: a nested list
    :return: a namedtuple SccdOutput
    """
    if len(sccd_pack_single) != defaults['SCCD']['PACK_ITEM']:
        raise Exception("the element number of sccd_pack_single must be {}".format(defaults['SCCD']['PACK_ITEM']))

    # convert to named tuple
    sccd_pack_single = SccdOutput(*sccd_pack_single)

    # replace the element to structured array
    if len(sccd_pack_single.rec_cg) == 0:
        sccd_pack_single = sccd_pack_single._replace(rec_cg=np.asarray(sccd_pack_single.rec_cg,
                                                                       dtype=np.float64))
    else:
        sccd_pack_single = sccd_pack_single._replace(rec_cg=np.asarray(sccd_pack_single.rec_cg,
                                                                       dtype=sccd_dt))
    if len(sccd_pack_single.nrt_model) > 0:
        sccd_pack_single = sccd_pack_single._replace(nrt_model=np.asarray(sccd_pack_single.nrt_model,
                                                                          dtype=nrtmodel_dt))
    if len(sccd_pack_single.nrt_queue) > 0:
        sccd_pack_single = sccd_pack_single._replace(nrt_queue=np.asarray(sccd_pack_single.nrt_queue,
                                                                          dtype=nrtqueue_dt))
    return sccd_pack_single


def save_1band_fromrefimage(array, out_path, ref_image_path=None, gtype=gdal.GDT_Int16):
    cols = array.shape[1]
    rows = array.shape[0]

    if ref_image_path is None:
        outdriver1 = gdal.GetDriverByName("GTiff")
        outdata = outdriver1.Create(out_path, rows, cols, 1, gtype)
        outdata.GetRasterBand(1).WriteArray(array)
        outdata.FlushCache()
    else:
        ref_image = gdal.Open(ref_image_path, gdal.GA_ReadOnly)
        trans = ref_image.GetGeoTransform()
        proj = ref_image.GetProjection()
        outdriver1 = gdal.GetDriverByName("GTiff")
        outdata = outdriver1.Create(out_path, rows, cols, 1, gtype)
        outdata.GetRasterBand(1).WriteArray(array)
        outdata.FlushCache()
        outdata.SetGeoTransform(trans)
        outdata.FlushCache()
        outdata.SetProjection(proj)
        outdata.FlushCache()
        del ref_image


def coefficient_matrix(dates, num_coefficients):
    """
    Fourier transform function to be used for the matrix of inputs for
    model fitting
    Args:
        dates: list of ordinal dates
        num_coefficients: how many coefficients to use to build the matrix
    Returns:
        Populated numpy array with coefficient values
    """
    slope_scale = 10000
    w23 = 2 * np.pi / 365.25
    matrix = np.zeros(shape=(num_coefficients), order='F')

    # lookup optimizations
    # Before optimization - 12.53% of total runtime
    # After optimization  - 10.57% of total runtime
    cos = np.cos
    sin = np.sin

    matrix[0] = 1
    matrix[1] = dates / slope_scale
    matrix[2] = cos(w23 * dates)
    matrix[3] = sin(w23 * dates)

    if num_coefficients >= 6:
        w45 = 2 * w23
        matrix[4] = cos(w45 * dates)
        matrix[5] = sin(w45 * dates)

    if num_coefficients >= 8:
        w67 = 3 * w23
        matrix[6] = cos(w67 * dates)
        matrix[7] = sin(w67 * dates)

    return matrix


def predict_ref(model, dates, num_coefficients=6):
    coef_matrix = coefficient_matrix(dates, num_coefficients)
    return np.dot(coef_matrix, model.T)


def generate_rowcolimage(ref_image_path, out_path):
    """
    a function to convert source image to index image (starting from 1, e.g., the first pixel is 10001)
    :param ref_image_path: the path for reference images
    :param out_path:  the path for output images
    :return:
    """
    # ref_image_path = '/home/coloury/Dropbox/UCONN/HLS/HLS.L30.T18TYM.2022074T153249.v2.0.B10.tif'
    ref_image = gdal.Open(ref_image_path, gdal.GA_ReadOnly)
    trans = ref_image.GetGeoTransform()
    proj = ref_image.GetProjection()
    cols = ref_image.RasterXSize
    rows = ref_image.RasterYSize

    i, j = np.indices(ref_image.ReadAsArray().shape[:2])
    index = (i + 1) * 10000 + j + 1

    outdriver1 = gdal.GetDriverByName("GTiff")
    outdata = outdriver1.Create(out_path, rows, cols, 1, gdal.GDT_Int32)
    outdata.GetRasterBand(1).WriteArray(index)
    outdata.FlushCache()
    outdata.SetGeoTransform(trans)
    outdata.FlushCache()
    outdata.SetProjection(proj)
    outdata.FlushCache()
    del ref_image



import numpy as np
from pycold.ob_analyst import segmentation_floodfill
from pycold.ob_analyst import object_analysis
from pycold.ob_analyst import ObjectAnalystHPC
import yaml
import shutil
import pathlib

# Use this file to determine where the resources are.  If for some reason we
# are do not have __file__ available (e.g.  copy/pasting in IPython) then
# assume we are in the repo reoot.
# TODO: should likely use pkg_resources instead
try:
    TEST_RESOURCE_DPATH = (pathlib.Path(__file__).parent / 'resources').resolve()
except NameError:
    TEST_RESOURCE_DPATH = pathlib.Path('tests/resources').resolve()

with open(TEST_RESOURCE_DPATH / 'test_config_obanalyst.yaml', 'r') as yaml_obj:
    test_config = yaml.safe_load(yaml_obj)

date = 730329

cm_array = np.load(TEST_RESOURCE_DPATH / 'cm_maps/CM_maps_730329_2000210.npy')
cm_array_l1 = np.load(TEST_RESOURCE_DPATH / 'cm_maps/CM_maps_730297_2000178.npy')
cm_date_array = np.load(TEST_RESOURCE_DPATH / 'cm_maps/CM_date_maps_730329_2000210.npy')
cm_array_l1_date = np.load(TEST_RESOURCE_DPATH / 'cm_maps/CM_date_maps_730297_2000178.npy')


def test_workflow():
    ob_analyst = ObjectAnalystHPC(test_config, starting_date=date,
                                  stack_path=TEST_RESOURCE_DPATH,
                                  result_path=TEST_RESOURCE_DPATH,
                                  thematic_path=TEST_RESOURCE_DPATH / 'feature_maps')
    ob_analyst.hpc_preparation()
    ob_analyst.obia_execute(date, method='slic')
    assert ob_analyst.is_finished_object_analysis(date_list=[date])
    shutil.rmtree(ob_analyst.obia_path)
    shutil.rmtree(ob_analyst.obcold_recg_path)


def test_segmentation():
    [object_map_s1, cm_date_array_updated, object_map_s2, s1_info] = segmentation_floodfill(
        cm_array, cm_date_array, cm_array_l1, cm_array_l1_date)
    assert len(np.unique(object_map_s1)) > 1
    assert len(np.unique(object_map_s2)) > 1


def test_object_analysis():
    [object_map_s1, cm_date_array_updated, object_map_s2, s1_info] = segmentation_floodfill(
        cm_array, cm_date_array, cm_array_l1, cm_array_l1_date)
    classification_map = np.load(TEST_RESOURCE_DPATH / 'feature_maps/yearlyclassification_1999.npy')
    change_map = object_analysis(object_map_s1, object_map_s2, s1_info, classification_map)
    # import matplotlib.pyplot as plt
    # plt.imshow(change_map)
    # test_config = yaml.safe_load(yaml_obj)
    assert len(np.unique(change_map)) > 1

    # import matplotlib
    # from random import random
    # import matplotlib.pyplot as plt
    # colors = [(1,1,1)] + [(random(),random(),random()) for i in range(255)]
    # new_map = matplotlib.colors.LinearSegmentedColormap.from_list('new_map', colors, N=256)
    # plt.imshow(object_map_s1, cmap=new_map)


def test_get_lastyear_cmap_fromdate():
    ob_analyst = ObjectAnalystHPC(test_config, starting_date=date - 366,
                                  stack_path=TEST_RESOURCE_DPATH,
                                  result_path=TEST_RESOURCE_DPATH,
                                  thematic_path=TEST_RESOURCE_DPATH / 'feature_maps')
    cmap = ob_analyst.get_lastyear_cmap_fromdate(date)
    assert cmap is not None



"""
Monkey patches for import-time behavior
"""
import warnings

warnings.filterwarnings("ignore")  # mainly to filter out lzma warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)



import os
import datetime as dt
from os.path import join, exists
import joblib
import time
import logging
import sys
import numpy as np
import pandas as pd
from sklearn.ensemble import RandomForestClassifier
from osgeo import gdal_array
from pycold.app import defaults
from pycold.utils import get_block_y, get_block_x, get_col_index, get_row_index, assemble_array


def extract_features(cold_plot, band, ordinal_day_list, nan_val, feature_outputs=['a0', 'a1', 'b1']):
    """
    generate features for classification based on a plot-based rec_cg and a list of days to be predicted
    Parameters
    ----------
    cold_plot: nested array
        plot-based rec_cg
    band: integer
        the predicted band number range from 0 to 6
    ordinal_day_list: list
        a list of days that this function will predict every days as a list as output
    nan_val: integer
        NA value assigned to the output
    feature_outputs: a list of outputted feature name
        it must be within [a0, c1, a1, b1,a2, b2, a3, b3, rmse]
    Returns
    -------
        feature: a list (length = n_feature) of 1-array [len(ordinal_day_list)]
    """
    features = [np.full(len(ordinal_day_list), nan_val, dtype=np.double) for x in range(len(feature_outputs))]
    for index, ordinal_day in enumerate(ordinal_day_list):
        # print(index)
        for idx, cold_curve in enumerate(cold_plot):
            if idx == len(cold_plot) - 1:
                max_days = cold_plot[idx]['t_end']
            else:
                max_days = cold_plot[idx + 1]['t_start']

            if cold_curve['t_start'] <= ordinal_day < max_days:
                for n, feature in enumerate(feature_outputs):
                    if feature == 'a0':
                        features[n][index] = cold_curve['coefs'][band][0] + cold_curve['coefs'][band][1] * \
                                             ordinal_day / defaults['COMMON']['SLOPE_SCALE']
                        if np.isnan(features[n][index]):
                            features[n][index] = 0
                    elif feature == 'c1':
                        features[n][index] = cold_curve['coefs'][band][1] / defaults['COMMON']['SLOPE_SCALE']
                        if np.isnan(features[n][index]):
                            features[n][index] = 0
                    elif feature == 'a1':
                        features[n][index] = cold_curve['coefs'][band][2]
                        if np.isnan(features[n][index]):
                            features[n][index] = 0
                    elif feature == 'b1':
                        features[n][index] = cold_curve['coefs'][band][3]
                        if np.isnan(features[n][index]):
                            features[n][index] = 0
                    elif feature == 'a2':
                        features[n][index] = cold_curve['coefs'][band][4]
                        if np.isnan(features[n][index]):
                            features[n][index] = 0
                    elif feature == 'b2':
                        features[n][index] = cold_curve['coefs'][band][5]
                        if np.isnan(features[n][index]):
                            features[n][index] = 0
                    elif feature == 'a3':
                        features[n][index] = cold_curve['coefs'][band][6]
                        if np.isnan(features[n][index]):
                            features[n][index] = 0
                    elif feature == 'b3':
                        features[n][index] = cold_curve['coefs'][band][7]
                        if np.isnan(features[n][index]):
                            features[n][index] = 0
                    elif feature == 'rmse':
                        features[n][index] = cold_curve['rmse'][band]
                        if np.isnan(features[n][index]):
                            features[n][index] = 0
                    else:
                        raise Exception('the outputted feature must be in [a0, c1, a1, b1,a2, b2, a3, b3, rmse]')
                break
    return features


def generate_sample_num(label, sample_parameters):
    """
    generate sample number for each land cover category using the method from 'Optimizing selection of training and
    auxiliary data for operational land cover classification for the LCMAP initiative'
    Args:
        label: an array of land-cover categories, i.e., map
        sample_parameters: obcold parameter structure
    Returns:

    """
    unique_category, unique_counts_category = np.unique(label[label <= sample_parameters['total_landcover_category']],
                                                        return_counts=True)
    counts_category = np.array([0] * sample_parameters['total_landcover_category'])
    for x in range(len(unique_counts_category)):
        counts_category[int(unique_category[x] - 1)] = unique_counts_category[x]
    percate_samples = np.array([round(x * sample_parameters['total_samples'] / sum(counts_category)) for x in
                                counts_category])
    percate_samples[percate_samples > sample_parameters['max_category_samples']] = \
        sample_parameters['max_category_samples']
    percate_samples[percate_samples < sample_parameters['min_category_samples']] = \
        sample_parameters['min_category_samples']
    percate_samples = np.minimum(percate_samples, counts_category)  # needs to check not exceed the total category pixels
    return percate_samples


def get_features(path):
    """
    Parameters
    ----------
    path:
    Returns: (block_width*block_height, total feature) array
    -------
        the feature temp array file for block_id and year
    """
    return np.load(path)


class PyClassifier:
    def __init__(self, config, feature_outputs=['a0', 'a1', 'b1'], logger=None, band_num=7):
        """
        Parameters
        ----------
        config: from config.yaml
        feature_outputs: a list of outputted feature name
            it must be within [a0, c1, a1, b1,a2, b2, a3, b3, rmse]
        """
        self.config = config
        self.config['block_width'] = int(self.config['n_cols'] / self.config['n_block_x'])
        self.config['block_height'] = int(self.config['n_rows'] / self.config['n_block_y'])
        self.config['n_blocks'] = self.config['n_block_x'] * self.config['n_block_y']
        for feature in feature_outputs:
            assert feature in ['a0', 'c1', 'a1', 'b1', 'a2', 'b2', 'a3', 'b3', 'rmse']
        self.n_features = band_num * len(feature_outputs)
        self.feature_outputs = feature_outputs
        if logger is None:
            logging.basicConfig(level=logging.DEBUG,
                                format='%(asctime)s |%(levelname)s| %(funcName)-15s| %(message)s',
                                stream=sys.stdout)
            self.logger = logging.getLogger(__name__)
        else:
            self.logger = logger
        self.band_num = band_num

    def predict_features(self, block_id, cold_block, year_list_to_predict, ismat=False):
        """
        Parameters
        ----------
        block_id: integer
            the block id
        cold_block: nested array of colddt datatype
            the block-based change records produced by cold algorithms
        year_list_to_predict:
            the list of classification years
            Note that the reason for not parsing cold_block to get year bounds is that the year ranges of blocks
            may vary from each other, so the year bounds are required to be defined from the tile level, not block level
            such as from 'starting_end_date.txt'
        ismat:boolean
            True -> the input is MATLAB-outputted change records

        Returns
        -------
        an array [len(year_list_to_predict), block_width*block_height, n_features]
        """
        block_features = np.full((len(year_list_to_predict),
                                  self.config['block_width'] * self.config['block_height'],
                                  self.n_features),
                                 defaults['COMMON']['NAN_VAL'], dtype=np.float32)
        ordinal_day_list = [pd.Timestamp.toordinal(dt.date(year, 7, 1)) for year
                            in year_list_to_predict]
        if len(cold_block) == 0:
            self.logger.warning('the rec_cg file for block_id={} has no records'.format(block_id))
            return block_features

        cold_block_split = np.split(cold_block, np.argwhere(np.diff(cold_block['pos']) != 0)[:, 0] + 1)
        for element in cold_block_split:
            # the relative column number in the block
            i_col = get_col_index(element[0]["pos"], self.config['n_cols'],
                                  get_block_x(block_id, self.config['n_block_x']),
                                  self.config['block_width'])
            i_row = get_row_index(element[0]["pos"], self.config['n_cols'],
                                  get_block_y(block_id, self.config['n_block_x']),
                                  self.config['block_height'])

            for band in range(self.band_num):
                feature_row = extract_features(element, band, ordinal_day_list, defaults['COMMON']['NAN_VAL'],
                                               feature_outputs=self.feature_outputs)
                for index in range(int(self.n_features / self.band_num)):
                    block_features[:, i_row * self.config['block_width'] + i_col,
                                   int(band * self.n_features / self.band_num) + index] \
                        = feature_row[index]

        return block_features

    def train_rfmodel(self, full_feature_array, label):
        """
        Parameters
        ----------
        full_feature_array: (nrows, ncols, feature_number) array
            feature array for the whole tile
        label: (nrows, ncols) array
            reference array for the whole tile
        Returns
        -------
            a sklearn random forest model
        """
        assert label.shape == (self.config['n_rows'], self.config['n_cols'])
        samplecount = generate_sample_num(label, defaults['CLASSIFIER'])
        index_list = []
        label_list = []
        for i in range(defaults['CLASSIFIER']['total_landcover_category']):
            index = np.argwhere(label == i + 1)
            np.random.seed(42)  # set random seed to reproduce the same result
            index_sample = index[np.random.choice(len(index), int(samplecount[i]), replace=False)]
            index_list.append(index_sample)
            label_list.append(np.array([i + 1] * len(index_sample)))
        index_list = np.vstack(index_list)
        label_list = np.hstack(label_list)
        feature_extraction = np.array([full_feature_array[tuple(x)] for x in index_list]).astype(np.float32)
        rf_model = RandomForestClassifier(random_state=42)
        rf_model.fit(feature_extraction, label_list)
        return rf_model

    def classification_block(self, rf_model, tmp_feature):
        """
        make classification for temp feature array for a single year, single block
        Parameters
        ----------
        rf_model: sklearn random forest model
        tmp_feature:
        Return: (block_height, block_width) array
        -------
            cmap, the feature temp array file for block_id and year
        """
        cmap = rf_model.predict(tmp_feature).reshape(self.config['block_height'], self.config['block_width'])
        mask = np.all(tmp_feature == defaults['COMMON']['NAN_VAL'], axis=1).reshape(self.config['block_height'],
                                                                                    self.config['block_width'])
        cmap[mask] = 255
        return cmap

    # def _assemble_covermap(self, blocklist_yearlyclass, year):
    #     full_yearlyclass_array = assemble_array(blocklist_yearlyclass, self.config['n_block_x'])
    #     if (full_yearlyclass_array.shape[1] != self.config['n_cols']) or (full_yearlyclass_array.shape[0] !=
    #                                                                       self.config['n_rows']):
    #         logger.error('The assemble category map is incomplete for {}'.format(year))
    #         return full_yearlyclass_array[:, :, 0]  # we only return the first channel


class PyClassifierHPC(PyClassifier):
    """
    this class adds IO functions based on the HPC environment for the base class
    """
    def __init__(self, config, record_path, band_num=7, year_list_to_predict=list(range(1982, 2022)),
                 tmp_path=None, output_path=None, feature_outputs=['a0', 'a1', 'b1'],
                 seedmap_path=None, rf_path=None, logger=None):
        """
        Parameters
        ----------
        config: configuration structure from config.yaml
        record_path: str
            the path that saves change records
        year_list_to_predict:
            the list of classification years
        tmp_path: string, default is None
            the path to save temporal folder, if None, will set /record_path/feature_maps
        output_path: string, default is None
            the path to save classification map output, if None, will set /record_path/feature_maps
        feature_outputs: a list of outputted feature name
            it must be within [a0, c1, a1, b1,a2, b2, a3, b3, rmse]
        seedmap_path: the path for the seed map to produce rf model
        rf_path: the path for existing random forest forest
        logger: the logger handler
        """
        try:
            self._check_inputs_thematic(config, record_path, tmp_path, seedmap_path, rf_path)
        except ValueError or FileExistsError as e:
            raise e

        self.config = config
        self.config['block_width'] = int(self.config['n_cols'] / self.config['n_block_x'])
        self.config['block_height'] = int(self.config['n_rows'] / self.config['n_block_y'])
        self.config['n_blocks'] = self.config['n_block_x'] * self.config['n_block_y']
        self.record_path = record_path
        for feature in feature_outputs:
            assert feature in ['a0', 'c1', 'a1', 'b1', 'a2', 'b2', 'a3', 'b3', 'rmse']
        self.feature_outputs = feature_outputs

        if tmp_path is None:
            self.tmp_path = join(record_path, 'feature_maps')  # default path
        else:
            self.tmp_path = tmp_path

        if output_path is None:
            self.output_path = join(record_path, 'feature_maps')  # default path
        else:
            self.output_path = tmp_path

        self.n_features = band_num * len(feature_outputs)

        self.year_list_to_predict = year_list_to_predict
        self.seedmap_path = seedmap_path
        if rf_path is None:
            self.rf_path = join(self.output_path, 'rf.model')  # default path
        else:
            self.rf_path = rf_path

        if logger is None:
            logging.basicConfig(level=logging.DEBUG,
                                format='%(asctime)s |%(levelname)s| %(funcName)-15s| %(message)s',
                                stream=sys.stdout)
            self.logger = logging.getLogger(__name__)
        else:
            self.logger = logger

        self.band_num = band_num

    @staticmethod
    def _check_inputs_thematic(config, record_path, tmp_path,  seedmap_path,
                                rf_path):
        if type(config['n_rows']) != int or config['n_rows'] < 0:
            raise ValueError('n_rows must be positive integer')
        if type(config['n_cols']) != int or config['n_cols'] < 0:
            raise ValueError('n_cols must be positive integer')
        if type(config['n_block_x']) != int or config['n_block_x'] < 0:
            raise ValueError('n_block_x must be positive integer')
        if type(config['n_block_y']) != int or config['n_block_y'] < 0:
            raise ValueError('n_block_y must be positive integer')
        if type(config['n_block_y']) != int or config['n_block_y'] < 0:
            raise ValueError('n_block_y must be positive integer')

        if os.path.isdir(record_path) is False:
            raise FileExistsError('No such directory: {}'.format(record_path))

        if seedmap_path is not None:
            if os.path.isfile(seedmap_path) is False:
                raise FileExistsError('No such file: {}'.format(seedmap_path))

        if rf_path is not None:
            if os.path.isfile(rf_path) is False:
                raise FileExistsError('No such file: {}'.format(rf_path))

    def _save_features(self, block_id, block_features):
        '''
        Parameters
        ----------
        block_id: integer
        block_features
        Returns
        -------

        '''
        for id, year in enumerate(self.year_list_to_predict):
            np.save(os.path.join(self.tmp_path, 'tmp_feature_year{}_block{}.npy').format(year, block_id),
                    block_features[id, :, :])

    def _is_finished_step1_predict_features(self):
        for iblock in range(self.config['n_blocks']):
            if not exists(join(self.tmp_path, 'tmp_step1_predict_{}_finished.txt'.format(iblock + 1))):
                return False
        return True

    @staticmethod
    def _save_rf_model(rf_model, rf_path):
        joblib.dump(rf_model, rf_path, compress=3)

    def _is_finished_step2_train_rfmodel(self):
        return exists(self.rf_path)

    def _get_rf_model(self):
        return joblib.load(self.rf_path)

    def _save_yearlyclassification_maps(self, block_id, year, cmap):
        outfile = join(self.tmp_path, 'tmp_yearlyclassification{}_block{}.npy'.format(year, block_id))
        np.save(outfile, cmap)

    def _is_finished_step3_classification(self):
        """
        :return: True or false
        """
        for iblock in range(self.config['n_blocks']):
            if not exists(join(self.tmp_path, 'tmp_step3_classification_{}_finished.txt'.format(iblock + 1))):
                return False
        return True

    def _save_covermaps(self, full_yearlyclass_array, year):
        np.save(join(self.output_path, 'yearlyclassification_{}.npy'.format(year)), full_yearlyclass_array)

    def _clean(self):
        tmp_yearlyclass_filenames = [file for file in os.listdir(self.tmp_path)
                                     if file.startswith('tmp_')]
        for file in tmp_yearlyclass_filenames:
            os.remove(join(self.tmp_path, file))

    def _get_fullclassification_forcertainyear(self, year):
        tmp_yearlyclass_filenames = [file for file in os.listdir(self.tmp_path)
                                     if file.startswith('tmp_yearlyclassification{}'.format(year))]

        # sort to guarantee order follows low to high rows
        tmp_yearlyclass_filenames.sort(key=lambda t: int(t[t.find('block') + 5: t.find('.npy')]))
        return [np.load(join(self.tmp_path, file)).reshape(self.config['block_height'],
                                                                        self.config['block_width'], 1)
                for file in tmp_yearlyclass_filenames]

    def get_fullfeature_forcertainyear(self, year):
        tmp_feature_filenames = [file for file in os.listdir(self.tmp_path)
                                 if file.startswith('tmp_feature_year{}'.format(year))]
        if len(tmp_feature_filenames) < self.config['n_blocks']:
            self.logger.warning('tmp features are incomplete! should have {}; but actually have {} feature images'.
                                format(self.config['n_blocks'], len(tmp_feature_filenames)))

        tmp_feature_filenames.sort(
            key=lambda t: int(t[t.find('block') + 5: t.find('.npy')]))  # sorted by row number

        return [np.load(join(self.tmp_path, file)).reshape(self.config['block_height'], self.config['block_width'],
                self.n_features) for file in tmp_feature_filenames]

        # full_feature_array = assemble_array(, self.config['n_block_x'])
        # if (full_feature_array.shape[1] != self.config['n_cols']) or (full_feature_array.shape[0] !=
        #                                                               self.config['n_rows']):
        #     logger.error('The feature image is incomplete for {}'.format(year))
        # return full_feature_array

    def hpc_preparation(self):
        if exists(self.tmp_path) is False:
            try:
                os.mkdir(self.tmp_path)
            except IOError as e:
                raise e

        if exists(self.output_path) is False:
            try:
                os.mkdir(self.output_path)
            except IOError as e:
                raise e

    def step1_feature_generation(self, block_id):
        cold_block = np.load(join(self.record_path, 'record_change_x{}_y{}_cold.npy').format(
                             get_block_x(block_id, self.config['n_block_x']),
                             get_block_y(block_id, self.config['n_block_x'])))
        block_features = self.predict_features(block_id, cold_block, self.year_list_to_predict)
        self._save_features(block_id, block_features)
        with open(join(self.tmp_path, 'tmp_step1_predict_{}_finished.txt'.format(block_id)), 'w'):
            pass

    def step2_train_rf(self, ref_year=None, rf_path=None):
        while not self._is_finished_step1_predict_features():
            time.sleep(5)

        if ref_year is None:
            ref_year = defaults['CLASSIFIER']['training_year']

        # if ref_year not in self.year_list_to_predict:
        #     raise Exception("Ref_year {} is not in year_list_to_predict {}. "
        #                     "PLease included it and re-run step1_feature_generation".format(ref_year,
        #                                                                                     self.year_list_to_predict))

        full_feature_array = assemble_array(self.get_fullfeature_forcertainyear(ref_year),
                                            self.config['n_block_x'])
        label_array = gdal_array.LoadFile(os.fspath(self.seedmap_path))
        rf_model = self.train_rfmodel(full_feature_array, label_array)
        if rf_path is None:
            self._save_rf_model(rf_model, self.rf_path)
        else:
            self._save_rf_model(rf_model, rf_path)

    def step3_classification(self, block_id):
        while not self._is_finished_step2_train_rfmodel():
            time.sleep(5)
        try:
            rf_model = self._get_rf_model()
        except IOError as e:
            raise ("Please double check your rf model file directory or generate random forest model first:"
                   " {}".format(e))

        for year in self.year_list_to_predict:
            tmp_feature_block = get_features(join(self.tmp_path, 'tmp_feature_year{}_block{}.npy'.format(year, block_id)))
            cmap = self.classification_block(rf_model, tmp_feature_block)
            self._save_yearlyclassification_maps(block_id, year, cmap)
        with open(join(self.tmp_path, 'tmp_step3_classification_{}_finished.txt'.format(block_id)), 'w'):
            pass

    def step3_classification_sccd(self, block_id):
        while not self._is_finished_step2_train_rfmodel():
            time.sleep(5)
        try:
            rf_model = self._get_rf_model()
        except IOError as e:
            raise ("Please double check your rf model file directory or generate random forest model first:"
                   " {}".format(e))

        # for year in self.year_list_to_predict:
        #     tmp_feature_block = get_features(join(self.tmp_path, 'tmp_feature_year{}_block{}.npy'.format(year,
        #                                                                                                 block_id)))
        #     cmap = self.classification_block(rf_model, tmp_feature_block)
        #     self._save_yearlyclassification_maps(block_id, year, cmap)

        tmp_feature_block = get_features(join(self.tmp_path, 'tmp_feature_now_block{}.npy'.format(block_id)))
        cmap = self.classification_block(rf_model, tmp_feature_block)
        self._save_yearlyclassification_maps(block_id, 'now', cmap)
        with open(join(self.tmp_path, 'tmp_step3_classification_{}_finished.txt'.format(block_id)), 'w'):
            pass

    def step4_assemble(self, clean=True):
        while not self._is_finished_step3_classification():
            time.sleep(5)
        for year in self.year_list_to_predict:
            full_yearlyclass_array = assemble_array(self._get_fullclassification_forcertainyear(year),
                                                    self.config['n_block_x'])[:, :, 0]
            self._save_covermaps(full_yearlyclass_array, year)
        if clean:
            self._clean()  # _clean all temp files

    def step4_assemble_sccd(self, clean=True):
        while not self._is_finished_step3_classification():
            time.sleep(5)

        full_yearlyclass_array = assemble_array(self._get_fullclassification_forcertainyear('now'),
                                                self.config['n_block_x'])[:, :, 0]
        self._save_covermaps(full_yearlyclass_array.astype(np.uint8), 'now')

        # for year in self.year_list_to_predict:
        #     full_yearlyclass_array = assemble_array(self._get_fullclassification_forcertainyear(year),
        #                                             self.config['n_block_x'])[:, :, 0]
        #     self._save_covermaps(full_yearlyclass_array, year)
        if clean:
            self._clean()  # _clean all temp files

    def is_finished_step4_assemble(self):
        for year in self.year_list_to_predict:
            if not os.path.exists(join(self.output_path, 'yearlyclassification_{}.npy'.format(year))):
                return False
        else:
            return True



""" Main bootstrap and configuration module for pycold.  Any module that
requires configuration or services should import app and obtain the
configuration or service from here.
app.py enables a very basic but sufficient form of loose coupling
by setting names of services & configuration once and allowing other modules
that require these services/information to obtain them by name rather than
directly importing or instantiating.
Module level constructs are only evaluated once in a Python application's
lifecycle, usually at the time of first import. This pattern is borrowed
from Flask.
source: https://github.com/repository-preservation/lcmap-pyclass/blob/develop/pyclass/app.py
"""
import yaml
from importlib import resources as importlib_resources


class Defaults(dict):
    def __init__(self, config_path='parameters.yaml'):
        with open(config_path, 'r') as f:
            super(Defaults, self).__init__(yaml.safe_load(f.read()))

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError('No such attribute: ' + name)

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        if name in self:
            del self[name]
        else:
            raise AttributeError('No such attribute: ' + name)


with importlib_resources.path('pycold', 'constants.yaml') as const_fpath:
    # import pathlib
    # const_fpath = pathlib.Path(__file__).parent / 'constants.yaml'
    defaults = Defaults(const_fpath)



from pycold import sccd_detect, sccd_update
from pycold.utils import read_data
import pickle
import pathlib
from collections import namedtuple
SccdOutput = namedtuple("SccdOutput", "position rec_cg min_rmse nrt_mode nrt_model nrt_queue")

# Use this file to determine where the resources are.  If for some reason we
# are do not have __file__ available (e.g.  copy/pasting in IPython) then
# assume we are in the repo reoot.
# TODO: should likely use pkg_resources instead
try:
    TEST_RESOURCE_DPATH = (pathlib.Path(__file__).parent / 'resources').resolve()
except NameError:
    TEST_RESOURCE_DPATH = pathlib.Path('tests/resources').resolve()


def test_sccd_detect():
    import pytest
    pytest.skip("Disable SCCD for now")
    # running COLD for a Landsat time series provided by a csv
    in_path = TEST_RESOURCE_DPATH / 'spectral_336_3980_obs.csv'
    data = read_data(in_path)
    dates, blues, greens, reds, nirs, swir1s, swir2s, thermals, qas, sensor = data.copy()  # exclude header
    sccd_result = sccd_detect(dates, blues, greens, reds, nirs, swir1s, swir2s, thermals, qas)
    assert sccd_result.rec_cg['t_break'] == 735016
    assert sccd_result.rec_cg['num_obs'] == 141
    assert sccd_result.nrt_model['num_obs'] == 63
    assert sccd_result.nrt_mode == 1


def test_sccd_update():
    import pytest
    pytest.skip("Disable SCCD for now")
    import numpy as np
    pickle_file = open(TEST_RESOURCE_DPATH / 'spectral_15043028_sccdplot.pickle', 'rb')

    sccd_plot = SccdOutput(*pickle.load(pickle_file))

    # Test that the named tuple structure is the same
    from pycold._pycold_cython import SccdOutput as SccdOutputCy
    assert SccdOutputCy._fields == SccdOutput._fields
    assert sccd_plot._fields == SccdOutput._fields

    in_path = TEST_RESOURCE_DPATH / 'spectral_15043028_ext.csv'
    data = read_data(in_path)
    dates, blues, greens, reds, nirs, swir1s, swir2s, thermals, qas, sensor = data.copy()  # exclude header

    for idx, (key, item) in enumerate(zip(sccd_plot._fields, sccd_plot)):
        print('----------')
        print(f'idx={idx}, key={key}')
        if isinstance(item, np.ndarray):
            item.shape
            print(f'item.shape={item.shape}')
            print(f'item.dtype={item.dtype}')
        else:
            print(f'item={item}')

    # Check that the packed numpy dtype hasn't unexpectedly changed between
    # testdata and the current library version.
    from pycold._pycold_cython import nrtmodel_dt
    sccd_pack = sccd_plot
    assert nrtmodel_dt.itemsize == sccd_pack.nrt_model.dtype.itemsize, (
        'The test data does not correspond to the current code')

    sccd_plot_new = sccd_update(sccd_pack, dates, blues, greens, reds, nirs, swir1s, swir2s, thermals, qas)
    assert sccd_plot_new.nrt_model['num_obs'] == 64
    assert sccd_plot_new.nrt_model['obs_date_since1982'][0][3] == 9059



"""
PyCOLD setup script
"""

__notes__ = """
# Check contents of wheel
rm -rf _skbuild/ dist/ tool/python/pycold.egg-info/
python setup.py bdist_wheel && unzip -l dist/pycold-0.1.0-cp38-cp38-linux_x86_64.whl

pip install dist/pycold-0.1.0-cp38-cp38-linux_x86_64.whl
python -c "import pycold"
"""
import sys
import os
from setuptools import find_packages

# Allow disabling of C-extensions to debug pure-python code
PYCOLD_DISABLE_C_EXTENSIONS = os.environ.get('PYCOLD_DISABLE_C_EXTENSIONS', '')
if PYCOLD_DISABLE_C_EXTENSIONS == '1':
    from setuptools import setup
else:
    from skbuild import setup


def parse_version(fpath):
    """
    Statically parse the version number from a python file
    """
    import ast
    from os.path import exists
    if not exists(fpath):
        raise ValueError('fpath={!r} does not exist'.format(fpath))
    with open(fpath, 'r') as file_:
        sourcecode = file_.read()
    pt = ast.parse(sourcecode)
    class VersionVisitor(ast.NodeVisitor):
        def visit_Assign(self, node):
            for target in node.targets:
                if getattr(target, 'id', None) == '__version__':
                    self.version = node.value.s
    visitor = VersionVisitor()
    visitor.visit(pt)
    return visitor.version


def parse_requirements(fname='requirements.txt', versions=False):
    """
    Parse the package dependencies listed in a requirements file but strips
    specific versioning information.

    Args:
        fname (str): path to requirements file
        versions (bool | str, default=False):
            If true include version specs.
            If strict, then pin to the minimum version.

    Returns:
        List[str]: list of requirements items
    """
    from os.path import exists, dirname, join
    import re
    require_fpath = fname

    def parse_line(line, dpath=''):
        """
        Parse information from a line in a requirements text file

        line = 'git+https://a.com/somedep@sometag#egg=SomeDep'
        line = '-e git+https://a.com/somedep@sometag#egg=SomeDep'
        """
        # Remove inline comments
        comment_pos = line.find(' #')
        if comment_pos > -1:
            line = line[:comment_pos]

        if line.startswith('-r '):
            # Allow specifying requirements in other files
            target = join(dpath, line.split(' ')[1])
            for info in parse_require_file(target):
                yield info
        else:
            # See: https://www.python.org/dev/peps/pep-0508/
            info = {'line': line}
            if line.startswith('-e '):
                info['package'] = line.split('#egg=')[1]
            else:
                if ';' in line:
                    pkgpart, platpart = line.split(';')
                    # Handle platform specific dependencies
                    # setuptools.readthedocs.io/en/latest/setuptools.html
                    # #declaring-platform-specific-dependencies
                    plat_deps = platpart.strip()
                    info['platform_deps'] = plat_deps
                else:
                    pkgpart = line
                    platpart = None

                # Remove versioning from the package
                pat = '(' + '|'.join(['>=', '==', '>']) + ')'
                parts = re.split(pat, pkgpart, maxsplit=1)
                parts = [p.strip() for p in parts]

                info['package'] = parts[0]
                if len(parts) > 1:
                    op, rest = parts[1:]
                    version = rest  # NOQA
                    info['version'] = (op, version)
            yield info

    def parse_require_file(fpath):
        dpath = dirname(fpath)
        with open(fpath, 'r') as f:
            for line in f.readlines():
                line = line.strip()
                if line and not line.startswith('#'):
                    for info in parse_line(line, dpath=dpath):
                        yield info

    def gen_packages_items():
        if exists(require_fpath):
            for info in parse_require_file(require_fpath):
                parts = [info['package']]
                if versions and 'version' in info:
                    if versions == 'strict':
                        # In strict mode, we pin to the minimum version
                        if info['version']:
                            # Only replace the first >= instance
                            verstr = ''.join(info['version']).replace('>=', '==', 1)
                            parts.append(verstr)
                    else:
                        parts.extend(info['version'])
                if not sys.version.startswith('3.4'):
                    # apparently package_deps are broken in 3.4
                    plat_deps = info.get('platform_deps')
                    if plat_deps is not None:
                        parts.append(';' + plat_deps)
                item = ''.join(parts)
                yield item

    packages = list(gen_packages_items())
    return packages


def parse_description():
    """
    Parse the description in the README file
    """
    from os.path import dirname, join, exists
    readme_fpath = join(dirname(__file__), 'README.rst')
    # This breaks on pip install, so check that it exists.
    if exists(readme_fpath):
        with open(readme_fpath, 'r') as f:
            text = f.read()
        return text
    return ''


VERSION = parse_version('src/python/pycold/__init__.py')  # needs to be a global var for git tags

if __name__ == '__main__':

    # References:
    # https://stackoverflow.com/questions/19602582/pip-install-editable-links-to-wrong-path
    packages = find_packages('./src/python')

    setupkw = {}
    setupkw["classifiers"] = [
        # List of classifiers available at:
        # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
        # 'Operating System :: Microsoft :: Windows',  # TODO
        # 'Operating System :: MacOS',  # TODO
        'Operating System :: POSIX :: Linux',
        'License :: OSI Approved :: Apache Software License',
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ]
    setupkw['license'] = 'Apache 2',

    setup(
        package_dir={
            '': 'src/python',
        },
        name='pycold',
        url='https://github.com/GERSL/pycold',
        version=VERSION,
        description='python implementation of COntinuous monitoring of Land disturbances algorithm',
        install_requires=parse_requirements('requirements/runtime.txt'),
        long_description=parse_description(),
        long_description_content_type='text/x-rst',
        extras_require={
            'all': parse_requirements('requirements.txt'),
            'tests': parse_requirements('requirements/tests.txt'),
            'build': parse_requirements('requirements/build.txt'),
            'optional': parse_requirements('requirements/optional.txt'),
            'headless': parse_requirements('requirements/headless.txt'),
            'graphics': parse_requirements('requirements/graphics.txt'),
        },
        author='Su Ye',
        author_email='remotesensingsuy@gmail.com',
        packages=packages,
        include_package_data=True,
        package_data={
            'pycold': [
                'ob_parameters.yaml',
                'constants.yaml',
            ],
            'pycold.imagetool': [
                'config.yaml',
                'singlepath_landsat_conus.tif',
            ],
        },
        python_requires='>=3.8',
    )



# Author: Su Ye
# generating yearly, recent and first-disturbance maps from change records
import os
import numpy as np
import pandas as pd
from osgeo import gdal
import click
import datetime as datetime
from mpi4py import MPI
from osgeo import gdal_array
import pickle
import yaml
from collections import namedtuple

PACK_ITEM = 6
SccdOutput = namedtuple("SccdOutput", "position rec_cg min_rmse nrt_mode nrt_model nrt_queue")
output_sccd = np.dtype([('t_start', np.int32), ('t_break', np.int32), ('num_obs', np.int32),
                        ('coefs', np.float32, (6, 6)), ('rmse', np.float32, 6), ('magnitude', np.float32, 6)], align=True)

output_nrtqueue = np.dtype([('clry', np.short, 6), ('clrx_since1982', np.short)], align=True)
output_nrtmodel = np.dtype([('t_start_since1982', np.short), ('num_obs', np.short), ('obs', np.short, (6, 5)),
                            ('obs_date_since1982', np.short, 5), ('covariance', np.float32, (6, 36)),
                            ('nrt_coefs', np.float32, (6, 6)), ('H', np.float32, 6), ('rmse_sum', np.uint32, 6),
                            ('cm_outputs', np.short), ('cm_outputs_date', np.short)], align=True)


# copy from /pycold/src/python/pycold/pyclassifier.py because MPI has conflicts with the pycold package in UCONN HPC.
# Dirty approach!
def extract_features(cold_plot, band, ordinal_day_list, nan_val, feature_outputs=['a0', 'a1', 'b1']):
    """
    generate features for classification based on a plot-based rec_cg and a list of days to be predicted
    Parameters
    ----------
    cold_plot: nested array
        plot-based rec_cg
    band: integer
        the predicted band number range from 0 to 6
    ordinal_day_list: list
        a list of days that this function will predict every days as a list as output
    nan_val: integer
        NA value assigned to the output
    feature_outputs: a list of outputted feature name
        it must be within [a0, c1, a1, b1,a2, b2, a3, b3, rmse]
    Returns
    -------
        feature: a list (length = n_feature) of 1-array [len(ordinal_day_list)]
    """
    features = [np.full(len(ordinal_day_list), nan_val, dtype=np.double) for x in range(len(feature_outputs))]
    for index, ordinal_day in enumerate(ordinal_day_list):
        # print(index)
        for idx, cold_curve in enumerate(cold_plot):
            if idx == len(cold_plot) - 1:
                max_days = cold_plot[idx]['t_end']
            else:
                max_days = cold_plot[idx + 1]['t_start']

            if cold_curve['t_start'] <= ordinal_day < max_days:
                for n, feature in enumerate(feature_outputs):
                    if feature == 'a0':
                        features[n][index] = cold_curve['coefs'][band][0] + cold_curve['coefs'][band][1] * \
                                             ordinal_day / 100
                        if np.isnan(features[n][index]):
                            features[n][index] = 0
                    elif feature == 'c1':
                        features[n][index] = cold_curve['coefs'][band][1] / 100
                        if np.isnan(features[n][index]):
                            features[n][index] = 0
                    elif feature == 'a1':
                        features[n][index] = cold_curve['coefs'][band][2]
                        if np.isnan(features[n][index]):
                            features[n][index] = 0
                    elif feature == 'b1':
                        features[n][index] = cold_curve['coefs'][band][3]
                        if np.isnan(features[n][index]):
                            features[n][index] = 0
                    elif feature == 'a2':
                        features[n][index] = cold_curve['coefs'][band][4]
                        if np.isnan(features[n][index]):
                            features[n][index] = 0
                    elif feature == 'b2':
                        features[n][index] = cold_curve['coefs'][band][5]
                        if np.isnan(features[n][index]):
                            features[n][index] = 0
                    elif feature == 'a3':
                        features[n][index] = cold_curve['coefs'][band][6]
                        if np.isnan(features[n][index]):
                            features[n][index] = 0
                    elif feature == 'b3':
                        features[n][index] = cold_curve['coefs'][band][7]
                        if np.isnan(features[n][index]):
                            features[n][index] = 0
                    elif feature == 'rmse':
                        features[n][index] = cold_curve['rmse'][band]
                        if np.isnan(features[n][index]):
                            features[n][index] = 0
                    else:
                        raise Exception('the outputted feature must be in [a0, c1, a1, b1,a2, b2, a3, b3, rmse]')
                break
    return features


def index_sccdpack(sccd_pack_single):
    """
    convert list of sccdpack to namedtuple to facilitate parse,
    :param sccd_pack_single: a nested list
    :return: a namedtuple SccdOutput
    """
    if len(sccd_pack_single) != PACK_ITEM:
        raise Exception("the element number of sccd_pack_single must be {}".format(PACK_ITEM))

    # convert to named tuple
    sccd_pack_single = SccdOutput(*sccd_pack_single)

    # replace the element to structured array
    if len(sccd_pack_single.rec_cg) == 0:
        sccd_pack_single = sccd_pack_single._replace(rec_cg=np.asarray(sccd_pack_single.rec_cg,
                                                                       dtype=np.float64))
    else:
        sccd_pack_single = sccd_pack_single._replace(rec_cg=np.asarray(sccd_pack_single.rec_cg,
                                                                       dtype=output_sccd))
    if len(sccd_pack_single.nrt_model) > 0:
        sccd_pack_single = sccd_pack_single._replace(nrt_model=np.asarray(sccd_pack_single.nrt_model,
                                                                          dtype=output_nrtmodel))
    if len(sccd_pack_single.nrt_queue) > 0:
        sccd_pack_single = sccd_pack_single._replace(nrt_queue=np.asarray(sccd_pack_single.nrt_queue,
                                                                          dtype=output_nrtqueue))
    return sccd_pack_single


def getcategory_cold(cold_plot, i_curve):
    t_c = -200
    if cold_plot[i_curve]['magnitude'][3] > t_c and cold_plot[i_curve]['magnitude'][2] < -t_c and \
            cold_plot[i_curve]['magnitude'][4] < -t_c:
        if cold_plot[i_curve + 1]['coefs'][3, 1] > np.abs(cold_plot[i_curve]['coefs'][3, 1]) and \
                cold_plot[i_curve + 1]['coefs'][2, 1] < -np.abs(cold_plot[i_curve]['coefs'][2, 1]) and \
                cold_plot[i_curve + 1]['coefs'][4, 1] < -np.abs(cold_plot[i_curve]['coefs'][4, 1]):
            return 3  # aforestation
        else:
            return 2  # regrowth
    else:
        return 1  # land disturbance


def getcategory_obcold(cold_plot, i_curve, last_dist_type):
    t_c = -250
    if cold_plot[i_curve]['magnitude'][3] > t_c and cold_plot[i_curve]['magnitude'][2] < -t_c and cold_plot[i_curve]['magnitude'][4] < -t_c:
        if cold_plot[i_curve + 1]['coefs'][3, 1] > np.abs(cold_plot[i_curve]['coefs'][3, 1]) and \
                cold_plot[i_curve + 1]['coefs'][2, 1] < -np.abs(cold_plot[i_curve]['coefs'][2, 1]) and \
                cold_plot[i_curve + 1]['coefs'][4, 1] < -np.abs(cold_plot[i_curve]['coefs'][4, 1]):
            return 3  # aforestation
        else:
            return 2  # regrowth
    else:
        if i_curve > 0:
            if (cold_plot[i_curve]['t_break'] - cold_plot[i_curve - 1]['t_break'] > 365.25 * 5) or (last_dist_type != 1):
                return 1
            flip_count = 0
            for b in range(5):
                if cold_plot[i_curve]['magnitude'][b + 1] * cold_plot[i_curve - 1]['magnitude'][b + 1] < 0:
                    flip_count = flip_count + 1
            if flip_count >= 4:
                return 4
            else:
                return 1
        else:
            return 1  # land disturbance


# for sccd we won't consider afforestation
def getcategory_sccd(cold_plot, i_curve):
    t_c = -200
    if cold_plot[i_curve]['magnitude'][3] > t_c and cold_plot[i_curve]['magnitude'][2] < -t_c and \
            cold_plot[i_curve]['magnitude'][4] < -t_c:
        return 2  # regrowth
    else:
        return 1  # land disturbance


@click.command()
@click.option('--reccg_path', type=str, help='rec_cg folder')
@click.option('--reference_path', type=str, help='image path used to provide georeference for output images')
@click.option('--out_path', type=str, help='output folder for saving image')
@click.option('--method', type=click.Choice(['COLD', 'OBCOLD', 'SCCDOFFLINE']), default='COLD', help='the algorithm used for processing')
@click.option('--yaml_path', type=str, help='path for yaml file')
@click.option('--year_lowbound', type=int, default=1982, help='the starting year for exporting')
@click.option('--year_uppbound', type=int, default=2020, help='the ending year for exporting')
@click.option('--coefs', default=False, help='if output coefs layers')
@click.option('--coefs_bands', default=[0, 1, 2, 3, 4, 5], help='indicate the bands for output coefs_bands,'
                                                                ' only works when coefs is True; note that the band order is'
                                                                'b,g,r,n, s1,s2, t')
def main(reccg_path, reference_path, out_path, method, year_lowbound, year_uppbound, yaml_path, coefs, coefs_bands):
    # reference_path = '/Users/coloury/Dropbox/UCONN/spatial/test_results/h016v010/recentdist_map_COLD.tif'
    # method = 'SCCDOFFLINE'
    # yaml_path = '/home/coloury/Dropbox/Documents/PyCharmProjects/HLS_NRT/config_hls.yaml'
    # reccg_path ='/home/coloury'
    # out_path = '/home/coloury'
    # year_lowbound = 1982
    # year_uppbound = 2020

    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    n_process = comm.Get_size()

    if method == 'OBCOLD':
        reccg_path = os.path.join(reccg_path, 'obcold')
        out_path = os.path.join(out_path, 'obcold_maps')
    elif method == 'COLD':
        out_path = os.path.join(out_path, 'cold_maps')
    elif method == 'SCCDOFFLINE':
        out_path = os.path.join(out_path, 'sccd_maps')

    # outname'obcold':
    # outname = 'breakyear_cold_h11v9_{}_{}_{}'.format(lower_year, upper_year, method)
    if method == 'SCCDOFFLINE':
        dt = np.dtype([('t_start', np.int32),
                       ('t_break', np.int32),
                       ('num_obs', np.int32),
                       ('coefs', np.float32, (6, 6)),  # note that the slope coefficient was scaled up by 10000
                       ('rmse', np.float32, 6),
                       ('magnitude', np.float32, 6)])
    else:
        dt = np.dtype([('t_start', np.int32),
                       ('t_end', np.int32),
                       ('t_break', np.int32),
                       ('pos', np.int32),
                       ('num_obs', np.int32),
                       ('category', np.short),
                       ('change_prob', np.short),
                       ('coefs', np.float32, (7, 8)),   # note that the slope coefficient was scaled up by 10000
                       ('rmse', np.float32, 7),
                       ('magnitude', np.float32, 7)])

    if rank == 0:
        if not os.path.exists(out_path):
            os.makedirs(out_path)

        ref_image = gdal.Open(reference_path, gdal.GA_ReadOnly)
        trans = ref_image.GetGeoTransform()
        proj = ref_image.GetProjection()
        cols = ref_image.RasterXSize
        rows = ref_image.RasterYSize

        with open(yaml_path, 'r') as yaml_obj:
            config = yaml.safe_load(yaml_obj)

        config['block_width'] = int(config['n_cols'] / config['n_block_x'])  # width of a block
        config['block_height'] = int(config['n_rows'] / config['n_block_y'])  # height of a block
        config['n_blocks'] = config['n_block_x'] * config['n_block_y']  # total number of blocks
    else:
        trans = None
        proj = None
        cols = None
        rows = None
        config = None

    trans = comm.bcast(trans, root=0)
    proj = comm.bcast(proj, root=0)
    cols = comm.bcast(cols, root=0)
    rows = comm.bcast(rows, root=0)
    config = comm.bcast(config, root=0)

    ranks_percore = int(np.ceil(config['n_blocks'] / n_process))
    for i in range(ranks_percore):
        iblock = n_process * i + rank
        if iblock >= config['n_blocks']:
            break
        current_block_y = int(np.floor(iblock / config['n_block_x'])) + 1
        current_block_x = iblock % config['n_block_x'] + 1
        if method == 'OBCOLD':
            filename = 'record_change_x{}_y{}_obcold.npy'.format(current_block_x, current_block_y)
        elif method == 'COLD':
            filename = 'record_change_x{}_y{}_cold.npy'.format(current_block_x, current_block_y)
        elif method == 'SCCDOFFLINE':
            filename = 'record_change_x{}_y{}_sccd.npy'.format(current_block_x, current_block_y)

        results_block = [np.full((config['block_height'], config['block_width']), -9999, dtype=np.int32)
                         for t in range(year_uppbound - year_lowbound + 1)]

        print('processing the rec_cg file {}'.format(os.path.join(reccg_path, filename)))
        if not os.path.exists(os.path.join(reccg_path, filename)):
            # reccg_path = '/home/coloury'
            print('the rec_cg file {} is missing'.format(os.path.join(reccg_path, filename)))
            for year in range(year_lowbound, year_uppbound + 1):
                outfile = os.path.join(out_path, 'tmp_map_block{}_{}.npy'.format(iblock + 1, year))
                np.save(outfile, results_block[year - year_lowbound])
            continue
        if method == 'SCCDOFFLINE':
            file = open(os.path.join(reccg_path, filename), 'rb')
            cold_block = []
            while True:
                try:
                    cold_block.append(index_sccdpack(pickle.load(file)))
                except EOFError:
                    break
            file.close()
        else:
            cold_block = np.array(np.load(os.path.join(reccg_path, filename)), dtype=dt)
        # cold_block = [np.array(element, dtype=dt) for element in cold_block]
        # if len(cold_block) == 0:
        #     print('the rec_cg file {} is missing'.format(dat_pth))
        #     for year in range(year_lowbound, year_uppbound+1):
        #         outfile = os.path.join(out_path, 'tmp_map_block{}_{}.npy'.format(iblock + 1, year))
        #         np.save(outfile, results_block[year - year_lowbound])
        #     continue
        if method == 'SCCDOFFLINE':
            for count, plot in enumerate(cold_block):
                for i_count, curve in enumerate(plot.rec_cg):
                    if curve['t_break'] == 0 or count == (len(cold_block) - 1):  # last segment
                        continue

                    i_col = int((plot.position  - 1) % config['n_cols']) - \
                            (current_block_x - 1) * config['block_width']
                    i_row = int((plot.position  - 1) / config['n_cols']) - \
                            (current_block_y - 1) * config['block_height']
                    if i_col < 0:
                        print('Processing {} failed: i_row={}; i_col={} for {}'.format(filename,
                                                                                       i_row, i_col, filename))
                        # return
                    current_dist_type = getcategory_sccd(plot.rec_cg, i_count)
                    break_year = pd.Timestamp.fromordinal(curve['t_break']).year
                    if break_year < year_lowbound or break_year > year_uppbound:
                        continue
                    results_block[break_year -
                                  year_lowbound][i_row][i_col] = current_dist_type * 1000 + curve['t_break'] - \
                                                                 (pd.Timestamp.toordinal(datetime.date(break_year, 1, 1))) + 1
        else:
            cold_block.sort(order='pos')
            current_processing_pos = cold_block[0]['pos']
            current_dist_type = 0
            for count, curve in enumerate(cold_block):
                if curve['pos'] != current_processing_pos:
                    current_processing_pos = curve['pos']
                    current_dist_type = 0

                if curve['change_prob'] < 100 or curve['t_break'] == 0 or count == (len(cold_block) - 1):  # last segment
                    continue

                i_col = int((curve["pos"] - 1) % config['n_cols']) - \
                        (current_block_x - 1) * config['block_width']
                i_row = int((curve["pos"] - 1) / config['n_cols']) - \
                        (current_block_y - 1) * config['block_height']
                if i_col < 0:
                    dat_pth = '?'
                    print('Processing {} failed: i_row={}; i_col={} for {}'.format(filename, i_row, i_col, dat_pth))
                    return

                if method == 'OBCOLD':
                    current_dist_type = getcategory_obcold(cold_block, count, current_dist_type)
                else:
                    current_dist_type = getcategory_cold(cold_block, count)
                break_year = pd.Timestamp.fromordinal(curve['t_break']).year
                if break_year < year_lowbound or break_year > year_uppbound:
                    continue
                results_block[break_year - year_lowbound][i_row][i_col] = current_dist_type * 1000 + curve['t_break'] - \
                    (pd.Timestamp.toordinal(datetime.date(break_year, 1, 1))) + 1
            # e.g., 1315 means that disturbance happens at doy of 315

        # save the temp dataset out
        for year in range(year_lowbound, year_uppbound + 1):
            outfile = os.path.join(out_path, 'tmp_map_block{}_{}.npy'.format(iblock + 1, year))
            np.save(outfile, results_block[year - year_lowbound])

    # wait for all processes
    comm.Barrier()

    if rank == 0:
        # assemble
        for year in range(year_lowbound, year_uppbound + 1):
            tmp_map_blocks = [np.load(os.path.join(out_path, 'tmp_map_block{}_{}.npy'.format(x + 1, year)))
                              for x in range(config['n_blocks'])]

            results = np.hstack(tmp_map_blocks)
            results = np.vstack(np.hsplit(results, config['n_block_x']))

            for x in range(config['n_blocks']):
                os.remove(os.path.join(out_path, 'tmp_map_block{}_{}.npy'.format(x + 1, year)))
            mode_string = str(year) + '_break_map'
            outname = '{}_{}.tif'.format(mode_string, method)
            outfile = os.path.join(out_path, outname)
            outdriver1 = gdal.GetDriverByName("GTiff")
            outdata = outdriver1.Create(outfile, rows, cols, 1, gdal.GDT_Int16)
            outdata.GetRasterBand(1).WriteArray(results)
            outdata.FlushCache()
            outdata.SetGeoTransform(trans)
            outdata.FlushCache()
            outdata.SetProjection(proj)
            outdata.FlushCache()

        # output recent disturbance year
        recent_dist = np.full((config['n_rows'], config['n_cols']), 0, dtype=np.int16)
        for year in range(year_lowbound, year_uppbound + 1):
            mode_string = str(year) + '_break_map'
            outname = '{}_{}.tif'.format(mode_string, method)
            breakmap = gdal_array.LoadFile(os.path.join(out_path, outname))
            recent_dist[(breakmap / 1000).astype(np.byte) == 1] = year
        mode_string = 'recent_disturbance_map'
        outname = '{}_{}.tif'.format(mode_string, method)
        outfile = os.path.join(out_path, outname)
        outdriver1 = gdal.GetDriverByName("GTiff")
        outdata = outdriver1.Create(outfile, rows, cols, 1, gdal.GDT_Int16)
        outdata.GetRasterBand(1).WriteArray(recent_dist)
        outdata.FlushCache()
        outdata.SetGeoTransform(trans)
        outdata.FlushCache()
        outdata.SetProjection(proj)
        outdata.FlushCache()

        first_dist = np.full((config['n_rows'], config['n_cols']), 0, dtype=np.int16)
        for year in range(year_uppbound, year_lowbound - 1, -1):
            mode_string = str(year) + '_break_map'
            outname = '{}_{}.tif'.format(mode_string, method)
            breakmap = gdal_array.LoadFile(os.path.join(out_path, outname))
            first_dist[(breakmap / 1000).astype(np.byte) == 1] = year
        mode_string = 'first_disturbance_map'
        outname = '{}_{}.tif'.format(mode_string, method)
        outfile = os.path.join(out_path, outname)
        outdriver1 = gdal.GetDriverByName("GTiff")
        outdata = outdriver1.Create(outfile, rows, cols, 1, gdal.GDT_Int16)
        outdata.GetRasterBand(1).WriteArray(first_dist)
        outdata.FlushCache()
        outdata.SetGeoTransform(trans)
        outdata.FlushCache()
        outdata.SetProjection(proj)
        outdata.FlushCache()


if __name__ == '__main__':
    main()



__version__ = '0.1.0'

# Ensure monkey patches run first
from . import monkey  # NOQA

from . import _pycold_cython  # NOQA

from ._pycold_cython import (  # NOQA
    test_func, cold_detect,
    obcold_reconstruct, sccd_detect, sccd_update)
